########################################################################
# File: Operation.py
# Date: 2012/07/24 12:12:05
########################################################################

"""
:mod: Operation

.. module: Operation
  :synopsis: Operation implementation

Operation implementation
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Disable invalid names warning
# pylint: disable=invalid-name
__RCSID__ = "$Id$"

import datetime
import json
import six

# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.private.JSONUtils import RMSEncoder


########################################################################
class Operation(object):
    """
    :param int OperationID: OperationID as read from DB backend
    :param int RequestID: parent RequestID
    :param str Status: execution status
    :param str Type: operation to perform
    :param str Arguments: additional arguments
    :param str SourceSE: source SE name
    :param str TargetSE: target SE names as comma separated list
    :param str Catalog: catalog to use as comma separated list
    :param str Error: error string if any
    :param Request.Request parent: parent Request instance


    It is managed by SQLAlchemy, so the RequestID, OperationID should never be set by hand
    (except when constructed from JSON of course...)
    In principle, the _parent attribute could be totally managed by SQLAlchemy. However, it is
    set only when inserted into the DB, this is why I manually set it in the Request _notify

    """

    # # max files in a single operation
    MAX_FILES = 10000

    # # all states
    ALL_STATES = ("Queued", "Waiting", "Scheduled", "Assigned", "Failed", "Done", "Canceled")
    # # final states
    FINAL_STATES = ("Failed", "Done", "Canceled")
    # # valid attributes
    ATTRIBUTE_NAMES = [
        "OperationID",
        "RequestID",
        "Type",
        "Status",
        "Arguments",
        "Order",
        "SourceSE",
        "TargetSE",
        "Catalog",
        "Error",
        "CreationTime",
        "SubmitTime",
        "LastUpdate",
    ]

    _datetimeFormat = "%Y-%m-%d %H:%M:%S"

    def __init__(self, fromDict=None):
        """c'tor

        :param self: self reference
        :param dict fromDict: attributes dictionary
        """
        self._parent = None

        now = datetime.datetime.utcnow().replace(microsecond=0)
        self._SubmitTime = now
        self._LastUpdate = now
        self._CreationTime = now

        self._Status = "Queued"
        self._Order = 0
        self.__files__ = []

        self.TargetSE = None
        self.SourceSE = None
        self._Arguments = None
        self.Error = None
        self.Type = None
        self._Catalog = None

        if isinstance(fromDict, six.string_types):
            fromDict = json.loads(fromDict)
        elif not isinstance(fromDict, dict):
            fromDict = {}

        if "Files" in fromDict:
            for fileDict in fromDict.get("Files", []):
                self.addFile(File(fileDict))

            del fromDict["Files"]

        for key, value in fromDict.items():
            # The JSON module forces the use of UTF-8, which is not properly
            # taken into account in DIRAC.
            # One would need to replace all the '== str' with 'in six.string_types'
            # This is converting `unicode` to `str` and doesn't make sense in Python 3
            if six.PY2 and isinstance(value, six.string_types):
                value = value.encode()
            if value:
                setattr(self, key, value)

    # # protected methods for parent only
    def _notify(self):
        """notify self about file status change"""
        fStatus = set(self.fileStatusList())
        if fStatus == set(["Failed"]):
            # All files Failed -> Failed
            newStatus = "Failed"
        elif "Scheduled" in fStatus:
            newStatus = "Scheduled"
        elif "Waiting" in fStatus:
            newStatus = "Queued"
        elif "Failed" in fStatus:
            newStatus = "Failed"
        else:
            self.Error = ""
            newStatus = "Done"

        # If the status moved to Failed or Done, update the lastUpdate time
        if newStatus in ("Failed", "Done", "Scheduled"):
            if self._Status != newStatus:
                self._LastUpdate = datetime.datetime.utcnow().replace(microsecond=0)

        self._Status = newStatus
        if self._parent:
            self._parent._notify()

    def _setQueued(self, caller):
        """don't touch"""
        if caller == self._parent:
            self._Status = "Queued"

    def _setWaiting(self, caller):
        """don't touch as well"""
        if caller == self._parent:
            self._Status = "Waiting"

    # # Files arithmetics
    def __contains__(self, opFile):
        """in operator"""
        return opFile in self.__files__

    def __iadd__(self, opFile):
        """+= operator"""
        if len(self) >= Operation.MAX_FILES:
            raise RuntimeError("too many Files in a single Operation")
        self.addFile(opFile)
        return self

    def addFile(self, opFile):
        """add :opFile: to operation

        .. warning::

          You cannot add a File object that has already been added to another operation. They must be different objects
        """
        if len(self) >= Operation.MAX_FILES:
            raise RuntimeError("too many Files in a single Operation")
        if opFile not in self:
            self.__files__.append(opFile)
            opFile._parent = self
        self._notify()

    # # helpers for looping
    def __iter__(self):
        """files iterator"""
        return self.__files__.__iter__()

    def __getitem__(self, i):
        """[] op for opFiles"""
        return self.__files__.__getitem__(i)

    def __delitem__(self, i):
        """remove file from op, only if OperationID is NOT set"""
        self.__files__.__delitem__(i)
        self._notify()

    def __setitem__(self, i, opFile):
        """overwrite opFile"""
        self.__files__.__setitem__(i, opFile)
        opFile._parent = self
        self._notify()

    def fileStatusList(self):
        """get list of files statuses"""
        return [subFile.Status for subFile in self]

    def __bool__(self):
        """for comparisons"""
        return True

    # For Python 2 compatibility
    __nonzero__ = __bool__

    def __len__(self):
        """nb of subFiles"""
        return len(self.__files__)

    @property
    def sourceSEList(self):
        """helper property returning source SEs as a list"""
        return self.SourceSE.split(",") if self.SourceSE else [""]

    @property
    def targetSEList(self):
        """helper property returning target SEs as a list"""
        return self.TargetSE.split(",") if self.TargetSE else [""]

    @property
    def Arguments(self):
        return self._Arguments

    @Arguments.setter
    def Arguments(self, value):
        if isinstance(value, six.text_type):
            value = value.encode()
        if not isinstance(value, bytes):
            raise TypeError("Arguments should be bytes!")
        self._Arguments = value

    @property
    def Catalog(self):
        """catalog prop"""
        return self._Catalog

    @Catalog.setter
    def Catalog(self, value):
        """catalog setter"""
        if not isinstance(value, six.string_types + (list,)):
            raise TypeError("wrong type for value")
        if isinstance(value, six.string_types):
            value = value.split(",")

        value = ",".join(list(set([str(item).strip() for item in value if str(item).strip()])))

        if len(value) > 255:
            raise ValueError("Catalog list too long")
        self._Catalog = value if value else ""

    @property
    def catalogList(self):
        """helper property returning catalogs as list"""
        return self._Catalog.split(",") if self._Catalog else []

    @property
    def Status(self):
        """Status prop"""
        return self._Status

    @Status.setter
    def Status(self, value):
        """Status setter"""
        if value not in Operation.ALL_STATES:
            raise ValueError("unknown Status '%s'" % str(value))
        if self.__files__:
            self._notify()
        else:
            # If the status moved to Failed or Done, update the lastUpdate time
            if value in ("Failed", "Done"):
                if self._Status != value:
                    self._LastUpdate = datetime.datetime.utcnow().replace(microsecond=0)

            self._Status = value
            if self._parent:
                self._parent._notify()
        if self._Status == "Done":
            self.Error = ""

    @property
    def Order(self):
        """order prop"""
        if self._parent:
            self._Order = self._parent.indexOf(self) if self._parent else -1
        return self._Order

    @Order.setter
    def Order(self, value):
        """order prop"""
        self._Order = value

    @property
    def CreationTime(self):
        """operation creation time prop"""
        return self._CreationTime

    @CreationTime.setter
    def CreationTime(self, value=None):
        """creation time setter"""
        if not isinstance(value, (datetime.datetime,) + six.string_types):
            raise TypeError("CreationTime should be a datetime.datetime!")
        if isinstance(value, six.string_types):
            value = datetime.datetime.strptime(value.split(".")[0], self._datetimeFormat)
        self._CreationTime = value

    @property
    def SubmitTime(self):
        """subrequest's submit time prop"""
        return self._SubmitTime

    @SubmitTime.setter
    def SubmitTime(self, value=None):
        """submit time setter"""
        if not isinstance(value, (datetime.datetime,) + six.string_types):
            raise TypeError("SubmitTime should be a datetime.datetime!")
        if isinstance(value, six.string_types):
            value = datetime.datetime.strptime(value.split(".")[0], self._datetimeFormat)
        self._SubmitTime = value

    @property
    def LastUpdate(self):
        """last update prop"""
        return self._LastUpdate

    @LastUpdate.setter
    def LastUpdate(self, value=None):
        """last update setter"""
        if not isinstance(value, (datetime.datetime,) + six.string_types):
            raise TypeError("LastUpdate should be a datetime.datetime!")
        if isinstance(value, six.string_types):
            value = datetime.datetime.strptime(value.split(".")[0], self._datetimeFormat)
        self._LastUpdate = value
        if self._parent:
            self._parent.LastUpdate = value

    def __str__(self):
        """str operator"""
        return self.toJSON()["Value"]

    def toJSON(self):
        """Returns the JSON description string of the Operation"""
        try:
            jsonStr = json.dumps(self, cls=RMSEncoder)
            return S_OK(jsonStr)
        except Exception as e:
            return S_ERROR(str(e))

    def _getJSONData(self):
        """Returns the data that have to be serialized by JSON"""

        jsonData = {}

        for attrName in Operation.ATTRIBUTE_NAMES:

            # RequestID and OperationID might not be set since they are managed by SQLAlchemy
            if not hasattr(self, attrName):
                continue

            value = getattr(self, attrName)

            if isinstance(value, datetime.datetime):
                # We convert date time to a string
                jsonData[attrName] = value.strftime(self._datetimeFormat)  # pylint: disable=no-member
            else:
                jsonData[attrName] = value

        jsonData["Files"] = self.__files__

        return jsonData



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


class DirectoryExplorer(object):
    def __init__(self, sort=False, reverse=False):
        self.__toExplore = []
        self.__explored = set()
        self.__sort = sort
        self.__reverse = reverse

    def isActive(self):
        return len(self.__toExplore) > 0

    def getNumRemainingDirs(self):
        return len(self.__toExplore)

    def __popNextDir(self):
        if self.__reverse:
            return self.__toExplore.pop()[1]
        else:
            return self.__toExplore.pop(0)[1]

    def getNextDir(self):
        if self.__sort:
            self.__toExplore = sorted(self.__toExplore)
        try:
            nextDir = self.__popNextDir()
            while nextDir in self.__explored:
                nextDir = self.__popNextDir()
        except IndexError:
            return False
        self.__explored.add(nextDir)
        return nextDir

    def addDir(self, dirName, weight=None):
        if weight is None and self.__sort:
            weight = dirName.count("/")
        if dirName not in self.__explored:
            self.__toExplore.append((weight, dirName))

    def addDirList(self, dirList, weight=None):
        for dirName in dirList:
            self.addDir(dirName, weight)



""" Contains a class used for evaluating policies for accessing jobs/WMS/pilots accounting
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Security import Properties
from DIRAC.AccountingSystem.private.Policies.FilterExecutor import FilterExecutor


class JobPolicy(object):
    def __init__(self):
        self.__executor = FilterExecutor()
        self.__executor.addGlobalFilter(self.__checkConditions)

    def getListingConditions(self, credDict):
        condDict = {}
        userProps = credDict["properties"]

        if Properties.JOB_ADMINISTRATOR in userProps:
            return condDict
        elif Properties.JOB_MONITOR in userProps:
            return condDict
        elif Properties.ACCOUNTING_MONITOR in userProps:
            return condDict
        elif Properties.JOB_SHARING in userProps:
            condDict["UserGroup"] = [credDict["group"]]

        return condDict

    def checkRequest(self, iD, credDict, condDict, groupingList):
        return self.__executor.applyFilters(iD, credDict, condDict, groupingList)

    def __checkConditions(self, credDict, condDict, groupingField):
        userProps = credDict["properties"]

        if Properties.JOB_ADMINISTRATOR in userProps:
            return S_OK()
        elif Properties.JOB_MONITOR in userProps:
            return S_OK()
        elif Properties.ACCOUNTING_MONITOR in userProps:
            return S_OK()
        elif Properties.JOB_SHARING in userProps:
            if "User" in condDict:
                condDict["UserGroup"] = credDict["group"]
            if "User" == groupingField:
                condDict["UserGroup"] = credDict["group"]
            if "UserGroup" in condDict:
                condDict["UserGroup"] = credDict["group"]
            if "UserGroup" == groupingField:
                condDict["UserGroup"] = credDict["group"]
        elif Properties.NORMAL_USER in userProps:
            if "User" in condDict:
                condDict["User"] = credDict["username"]
            if "User" == groupingField:
                condDict["User"] = credDict["username"]
            if "UserGroup" in condDict:
                condDict["User"] = credDict["username"]
                condDict["UserGroup"] = credDict["group"]
            if "UserGroup" == groupingField:
                condDict["User"] = credDict["username"]
                condDict["UserGroup"] = credDict["group"]
            else:
                condDict["User"] = credDict["username"]
        else:
            if "User" in condDict:
                del condDict["User"]
            if "UserGroup" in condDict:
                del condDict["UserGroup"]
            if "User" == groupingField:
                return S_ERROR("You can't group plots by users!")
            if "UserGroup" == groupingField:
                return S_ERROR("You can't group plots by user groups!")

        return S_OK()

    def filterListingValues(self, credDict, dataDict):
        userProps = credDict["properties"]

        if Properties.JOB_ADMINISTRATOR in userProps:
            return S_OK(dataDict)
        elif Properties.JOB_MONITOR in userProps:
            return S_OK(dataDict)
        elif Properties.ACCOUNTING_MONITOR in userProps:
            return S_OK(dataDict)
        elif Properties.JOB_SHARING in userProps:
            return S_OK(dataDict)
        elif Properties.NORMAL_USER in userProps:
            return S_OK(dataDict)

        dataDict["User"] = []
        dataDict["UserGroup"] = []

        return S_OK(dataDict)



#!/usr/bin/env python
########################################################################
""" System Administrator Client Command Line Interface """

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys
import pprint
import os
import atexit
import readline
import datetime
import time

from DIRAC import gConfig, gLogger
from DIRAC.Core.Base.CLI import CLI, colorize
from DIRAC.FrameworkSystem.Client.SystemAdministratorClient import SystemAdministratorClient
from DIRAC.FrameworkSystem.Client.SystemAdministratorIntegrator import SystemAdministratorIntegrator
from DIRAC.FrameworkSystem.Client.ComponentMonitoringClient import ComponentMonitoringClient
from DIRAC.FrameworkSystem.Utilities import MonitoringUtilities
from DIRAC.MonitoringSystem.Client.MonitoringClient import MonitoringClient
from DIRAC.FrameworkSystem.Client.ComponentInstaller import gComponentInstaller
from DIRAC.Core.Utilities.Extensions import extensionsByPriority
from DIRAC.Core.Utilities import List
from DIRAC.Core.Utilities.PromptUser import promptUser
from DIRAC.Core.Utilities.PrettyPrint import printTable
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Security.ProxyInfo import getProxyInfo

__RCSID__ = "$Id$"


class SystemAdministratorClientCLI(CLI):
    """Line oriented command interpreter for administering DIRAC components"""

    def __init__(self, host=None):

        CLI.__init__(self)
        # Check if Port is given
        self.host = None
        self.port = None
        self.prompt = "[%s]> " % colorize("no host", "yellow")
        if host:
            self.__setHost(host)
        self.cwd = ""
        self.previous_cwd = ""
        self.homeDir = ""
        self.runitComponents = ["service", "agent", "executor", "consumer"]

        # store history
        histfilename = os.path.basename(sys.argv[0])
        historyFile = os.path.expanduser("~/.dirac/%s.history" % histfilename[0:-3])
        mkDir(os.path.dirname(historyFile))
        if os.path.isfile(historyFile):
            readline.read_history_file(historyFile)
        readline.set_history_length(1000)
        atexit.register(readline.write_history_file, historyFile)

    def __setHost(self, host):
        hostList = host.split(":")
        self.host = hostList[0]
        if len(hostList) == 2:
            self.port = hostList[1]
        else:
            self.port = None
        gLogger.notice("Pinging %s..." % self.host)
        result = self.__getClient().ping()
        if result["OK"]:
            colorHost = colorize(host, "green")
        else:
            self._errMsg("Could not connect to %s: %s" % (self.host, result["Message"]))
            colorHost = colorize(host, "red")
        self.prompt = "[%s]> " % colorHost

    def __getClient(self):
        return SystemAdministratorClient(self.host, self.port)

    def do_set(self, args):
        """
        Set options

        usage:

          set host <hostname>     - Set the hostname to work with
          set project <project>   - Set the project to install/upgrade in the host
        """
        if not args:
            gLogger.notice(self.do_set.__doc__)
            return

        cmds = {"host": (1, self.__do_set_host), "project": (1, self.__do_set_project)}

        args = List.fromChar(args, " ")

        for cmd in cmds:
            if cmd == args[0]:
                if len(args) != 1 + cmds[cmd][0]:
                    self._errMsg("Missing arguments")
                    gLogger.notice(self.do_set.__doc__)
                    return
                return cmds[cmd][1](args[1:])
        self._errMsg("Invalid command")
        gLogger.notice(self.do_set.__doc__)
        return

    def __do_set_host(self, args):
        host = args[0]
        if host.find(".") == -1 and host != "localhost":
            self._errMsg("Provide the full host name including its domain")
            return
        self.__setHost(host)

    def __do_set_project(self, args):
        project = args[0]
        result = self.__getClient().setProject(project)
        if not result["OK"]:
            self._errMsg("Cannot set project: %s" % result["Message"])
        else:
            gLogger.notice("Project set to %s" % project)

    def do_show(self, args):
        """
        Show list of components with various related information

        usage:

          show software      - show components for which software is available
          show installed     - show components installed in the host with runit system
          show setup         - show components set up for automatic running in the host
          show project       - show project to install or upgrade
          show status        - show status of the installed components
          show database      - show status of the databases
          show mysql         - show status of the MySQL server
          show log  <system> <service|agent> [nlines]
                             - show last <nlines> lines in the component log file
          show info          - show version of software and setup
          show doc <type> <system> <name>
                             - show documentation for a given service or agent
          show host          - show host related parameters
          show hosts         - show all available hosts
          show ports [host]  - show all ports used by a host. If no host is given, the host currently connected to is used
          show installations [ list | current | -n <Name> | -h <Host> | -s <System> | -m <Module> | -t <Type> | -itb <InstallationTime before>
                              | -ita <InstallationTime after> | -utb <UnInstallationTime before> | -uta <UnInstallationTime after> ]*
                             - show all the installations of components that match the given parameters
          show profile <system> <component> [ -s <size> | -h <host> | -id <initial date DD/MM/YYYY> | -it <initial time hh:mm>
                              | -ed <end date DD/MM/YYYY | -et <end time hh:mm> ]*
                             - show <size> log lines of profiling information for a component in the machine <host>
          show errors [*|<system> <service|agent>]
                             - show error count for the given component or all the components
                               in the last hour and day
        """

        argss = args.split()
        if not argss:
            gLogger.notice(self.do_show.__doc__)
            return

        option = argss[0]
        del argss[0]

        if option == "software":
            client = SystemAdministratorClient(self.host, self.port)
            result = client.getSoftwareComponents()
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                gLogger.notice("")
                pprint.pprint(result["Value"])
        elif option == "installed":
            client = SystemAdministratorClient(self.host, self.port)
            result = client.getInstalledComponents()
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                gLogger.notice("")
                pprint.pprint(result["Value"])
        elif option == "setup":
            client = SystemAdministratorClient(self.host, self.port)
            result = client.getSetupComponents()
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                gLogger.notice("")
                pprint.pprint(result["Value"])
        elif option == "project":
            result = SystemAdministratorClient(self.host, self.port).getProject()
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                gLogger.notice("Current project is %s" % result["Value"])
        elif option == "status":
            client = SystemAdministratorClient(self.host, self.port)
            result = client.getOverallStatus()
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                fields = ["System", "Name", "Module", "Type", "Setup", "Installed", "Runit", "Uptime", "PID"]
                records = []
                rDict = result["Value"]
                for compType in rDict:
                    for system in rDict[compType]:
                        components = sorted(rDict[compType][system])
                        for component in components:
                            record = []
                            if rDict[compType][system][component]["Installed"]:
                                module = str(rDict[compType][system][component]["Module"])
                                record += [system, component, module, compType.lower()[:-1]]
                                if rDict[compType][system][component]["Setup"]:
                                    record += ["Setup"]
                                else:
                                    record += ["NotSetup"]
                                if rDict[compType][system][component]["Installed"]:
                                    record += ["Installed"]
                                else:
                                    record += ["NotInstalled"]
                                record += [str(rDict[compType][system][component]["RunitStatus"])]
                                record += [str(rDict[compType][system][component]["Timeup"])]
                                record += [str(rDict[compType][system][component]["PID"])]
                                records.append(record)
                printTable(fields, records)
        elif option == "database" or option == "databases":
            client = SystemAdministratorClient(self.host, self.port)
            if not gComponentInstaller.mysqlPassword:
                gComponentInstaller.mysqlPassword = "LocalConfig"
            gComponentInstaller.getMySQLPasswords()
            result = client.getDatabases(gComponentInstaller.mysqlRootPwd)
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            resultSW = client.getAvailableDatabases()
            if not resultSW["OK"]:
                self._errMsg(resultSW["Message"])
                return

            sw = resultSW["Value"]
            installed = result["Value"]
            gLogger.notice("")
            for db in sw:
                if db in installed:
                    gLogger.notice(db.rjust(25), ": Installed")
                else:
                    gLogger.notice(db.rjust(25), ": Not installed")
            if not sw:
                gLogger.notice("No database found")
        elif option == "mysql":
            client = SystemAdministratorClient(self.host, self.port)
            result = client.getMySQLStatus()
            if not result["OK"]:
                self._errMsg(result["Message"])
            elif result["Value"]:
                gLogger.notice("")
                for par, value in result["Value"].items():
                    gLogger.notice((par.rjust(28), ":", value))
            else:
                gLogger.notice("No MySQL database found")
        elif option == "log":
            self.getLog(argss)
        elif option == "info":
            client = SystemAdministratorClient(self.host, self.port)
            result = client.getInfo()
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                gLogger.notice("")
                gLogger.notice("Setup:", result["Value"]["Setup"])
                for e, v in result["Value"]["Extensions"].items():
                    gLogger.notice("%s version" % e, v)
                gLogger.notice("")
        elif option == "host":
            client = SystemAdministratorClient(self.host, self.port)
            result = client.getHostInfo()
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                gLogger.notice("")
                gLogger.notice("Host info:")
                gLogger.notice("")

                fields = ["Parameter", "Value"]
                records = []
                for parameter in result["Value"].items():
                    if parameter[0] == "Extension":
                        extensions = parameter[1].split(",")
                        for extension in extensions:
                            extensionName, extensionVersion = extension.split(":")
                            records.append(["%sVersion" % extensionName, str(extensionVersion)])
                    else:
                        records.append([parameter[0], str(parameter[1])])

                printTable(fields, records)
        elif option == "hosts":
            client = ComponentMonitoringClient()
            result = client.getHosts({}, False, False)
            if not result["OK"]:
                self._errMsg("Error retrieving the list of hosts: %s" % (result["Message"]))
            else:
                hostList = result["Value"]
                gLogger.notice("")
                gLogger.notice(" " + "Host".center(32) + " " + "CPU".center(34) + " ")
                gLogger.notice(("-" * 69))
                for element in hostList:
                    gLogger.notice("|" + element["HostName"].center(32) + "|" + element["CPU"].center(34) + "|")
                gLogger.notice(("-" * 69))
                gLogger.notice("")
        elif option == "ports":
            if not argss:
                client = SystemAdministratorClient(self.host)
            else:
                hostname = argss[0]
                del argss[0]

                client = ComponentMonitoringClient()
                result = client.hostExists({"HostName": hostname})
                if not result["OK"]:
                    self._errMsg(result["Message"])
                    return
                else:
                    if not result["Value"]:
                        self._errMsg("Given host does not exist")
                        return

                client = SystemAdministratorClient(hostname)

            result = client.getUsedPorts()
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            pprint.pprint(result["Value"])
        elif option == "errors":
            self.getErrors(argss)
        elif option == "installations":
            self.getInstallations(argss)
        elif option == "doc":
            if len(argss) > 2:
                if argss[0] in ["service", "agent"]:
                    compType = argss[0]
                    compSystem = argss[1]
                    compModule = argss[2]
                    client = SystemAdministratorClient(self.host, self.port)
                    result = client.getComponentDocumentation(compType, compSystem, compModule)
                    if result["OK"]:
                        gLogger.notice(result["Value"])
                    else:
                        self._errMsg(result["Message"])
                else:
                    gLogger.notice(self.do_show.__doc__)
            else:
                gLogger.notice(self.do_show.__doc__)
        elif option == "profile":
            if len(argss) > 1:
                system = argss[0]
                del argss[0]
                component = argss[0]
                del argss[0]

                component = "%s_%s" % (system, component)

                argDict = {"-s": None, "-h": self.host, "-id": None, "-it": "00:00", "-ed": None, "-et": "00:00"}
                key = None
                for arg in argss:
                    if not key:
                        key = arg
                    else:
                        argDict[key] = arg
                        key = None

                size = None
                try:
                    if argDict["-s"]:
                        size = int(argDict["-s"])
                except ValueError:
                    self._errMsg("Argument 'size' must be an integer")
                    return
                host = argDict["-h"]
                initialDate = argDict["-id"]
                initialTime = argDict["-it"]
                endingDate = argDict["-ed"]
                endingTime = argDict["-et"]

                if initialDate:
                    initialDate = "%s %s" % (initialDate, initialTime)
                else:
                    initialDate = ""
                if endingDate:
                    endingDate = "%s %s" % (endingDate, endingTime)
                else:
                    endingDate = ""

                client = MonitoringClient()
                if size:
                    result = client.getLimitedData(host, component, size)
                else:
                    result = client.getDataForAGivenPeriod(host, component, initialDate, endingDate)

                if result["OK"]:
                    text = ""
                    headers = [list(result["Value"][0])]
                    for header in headers:
                        text += str(header).ljust(15)
                    gLogger.notice(text)
                    for record in result["Value"]:
                        for metric in record.values():
                            text += str(metric).ljust(15)
                        gLogger.notice(text)
                else:
                    self._errMsg(result["Message"])
            else:
                gLogger.notice(self.do_show.__doc__)
        else:
            gLogger.notice("Unknown option:", option)

    def getErrors(self, argss):
        """Get and gLogger.notice( out errors from the logs of specified components )"""
        component = ""
        if len(argss) < 1:
            component = "*"
        else:
            system = argss[0]
            if system == "*":
                component = "*"
            else:
                if len(argss) < 2:
                    gLogger.notice("")
                    gLogger.notice(self.do_show.__doc__)
                    return
                comp = argss[1]
                component = "/".join([system, comp])

        client = SystemAdministratorClient(self.host, self.port)
        result = client.checkComponentLog(component)
        if not result["OK"]:
            self._errMsg(result["Message"])
        else:
            fields = ["System", "Component", "Last hour", "Last day", "Last error"]
            records = []
            for cname in result["Value"]:
                system, component = cname.split("/")
                errors_1 = result["Value"][cname]["ErrorsHour"]
                errors_24 = result["Value"][cname]["ErrorsDay"]
                lastError = result["Value"][cname]["LastError"]
                lastError.strip()
                if len(lastError) > 80:
                    lastError = lastError[:80] + "..."
                records.append([system, component, str(errors_1), str(errors_24), lastError])
            records.sort()
            printTable(fields, records)

    def getInstallations(self, argss):
        """Get data from the component monitoring database"""
        display = "table"
        installationFilter = {}
        componentFilter = {}
        hostFilter = {}

        key = ""
        for arg in argss:
            if not key:
                if arg == "list":
                    display = "list"
                elif arg == "current":
                    installationFilter["UnInstallationTime"] = None
                elif arg == "-t":
                    key = "Component.Type"
                elif arg == "-m":
                    key = "Component.Module"
                elif arg == "-s":
                    key = "Component.System"
                elif arg == "-h":
                    key = "Host.HostName"
                elif arg == "-n":
                    key = "Instance"
                elif arg == "-itb":
                    key = "InstallationTime.smaller"
                elif arg == "-ita":
                    key = "InstallationTime.bigger"
                elif arg == "-utb":
                    key = "UnInstallationTime.smaller"
                elif arg == "-uta":
                    key = "UnInstallationTime.bigger"
            else:
                if "Component." in key:
                    componentFilter[key.replace("Component.", "")] = arg
                elif "Host." in key:
                    hostFilter[key.replace("Host.", "")] = arg
                else:
                    if "Time." in key:
                        arg = datetime.datetime.strptime(arg, "%d-%m-%Y")
                    installationFilter[key] = arg
                key = None

        client = ComponentMonitoringClient()
        result = client.getInstallations(installationFilter, componentFilter, hostFilter, True)
        if not result["OK"]:
            self._errMsg("Could not retrieve the installations: %s" % (result["Message"]))
            installations = None
        else:
            installations = result["Value"]

        if installations:
            if display == "table":
                gLogger.notice("")
                gLogger.notice(
                    " "
                    + "Num".center(5)
                    + " "
                    + "Host".center(20)
                    + " "
                    + "Name".center(20)
                    + " "
                    + "DIRACModule".center(20)
                    + " "
                    + "DIRACSystem".center(16)
                    + " "
                    + "Type".center(12)
                    + " "
                    + "Installed on".center(18)
                    + " "
                    + "Install by".center(12)
                    + " "
                    + "Uninstalled on".center(18)
                    + " "
                    + "Uninstall by".center(12)
                )
                gLogger.notice(("-") * 164)
                gLogger.notice(("-") * 164)
            for i, installation in enumerate(installations):
                if not installation["InstalledBy"]:
                    installedBy = ""
                else:
                    installedBy = installation["InstalledBy"]

                if not installation["UnInstalledBy"]:
                    uninstalledBy = ""
                else:
                    uninstalledBy = installation["UnInstalledBy"]

                if installation["UnInstallationTime"]:
                    uninstalledOn = installation["UnInstallationTime"].strftime("%d-%m-%Y %H:%M")
                    isInstalled = "No"
                else:
                    uninstalledOn = ""
                    isInstalled = "Yes"

                if display == "table":
                    gLogger.notice(
                        "|"
                        + str(i + 1).center(5)
                        + "|"
                        + installation["Host"]["HostName"].center(20)
                        + "|"
                        + installation["Instance"].center(20)
                        + "|"
                        + installation["Component"]["DIRACModule"].center(20)
                        + "|"
                        + installation["Component"]["DIRACSystem"].center(16)
                        + "|"
                        + installation["Component"]["Type"].center(12)
                        + "|"
                        + installation["InstallationTime"].strftime("%d-%m-%Y %H:%M").center(18)
                        + "|"
                        + installedBy.center(12)
                        + "|"
                        + uninstalledOn.center(18)
                        + "|"
                        + uninstalledBy.center(12)
                        + "|"
                    )
                    gLogger.notice(("-") * 164)
                elif display == "list":
                    gLogger.notice("")
                    gLogger.notice("Installation: ".rjust(20) + str(i + 1))
                    gLogger.notice("Installed: ".rjust(20) + isInstalled)
                    gLogger.notice("Host: ".rjust(20) + installation["Host"]["HostName"])
                    gLogger.notice("Name: ".rjust(20) + installation["Instance"])
                    gLogger.notice("Module: ".rjust(20) + installation["Component"]["DIRACModule"])
                    gLogger.notice("System: ".rjust(20) + installation["Component"]["DIRACSystem"])
                    gLogger.notice("Type: ".rjust(20) + installation["Component"]["Type"])
                    gLogger.notice(
                        "Installed on: ".rjust(20) + installation["InstallationTime"].strftime("%d-%m-%Y %H:%M")
                    )
                    if installedBy != "":
                        gLogger.notice("Installed by: ".rjust(20) + installedBy)
                    if uninstalledOn != "":
                        gLogger.notice("Uninstalled on: ".rjust(20) + uninstalledOn)
                        gLogger.notice("Uninstalled by: ".rjust(20) + uninstalledBy)
                else:
                    self._errMsg("No display mode was selected")
            gLogger.notice("")

    def getLog(self, argss):
        """Get the tail of the log file of the given component"""
        if len(argss) < 2:
            gLogger.notice("")
            gLogger.notice(self.do_show.__doc__)
            return

        system = argss[0]
        component = argss[1]
        nLines = 40
        if len(argss) > 2:
            nLines = int(argss[2])
        client = SystemAdministratorClient(self.host, self.port)
        result = client.getLogTail(system, component, nLines)
        if not result["OK"]:
            self._errMsg(result["Message"])
        elif result["Value"]:
            for line in result["Value"]["_".join([system, component])].split("\n"):
                gLogger.notice("   ", line)

        else:
            gLogger.notice("No logs found")

    def do_install(self, args):
        """
        Install various DIRAC components

        usage:

          install db <databaseName>
          install service <system> <service> [-m <ModuleName>] [-p <Option>=<Value>] [-p <Option>=<Value>] ...
          install agent <system> <agent> [-m <ModuleName>] [-p <Option>=<Value>] [-p <Option>=<Value>] ...
          install executor <system> <executor> [-m <ModuleName>] [-p <Option>=<Value>] [-p <Option>=<Value>] ...
        """
        argss = args.split()
        if not argss:
            gLogger.notice(self.do_install.__doc__)
            return

        option = argss[0]
        del argss[0]

        # Databases
        if option == "db":
            if not argss:
                gLogger.notice(self.do_install.__doc__)
                return
            database = argss[0]
            client = SystemAdministratorClient(self.host, self.port)

            result = client.getAvailableDatabases()
            if not result["OK"]:
                self._errMsg("Can not get database list: %s" % result["Message"])
                return
            if database not in result["Value"]:
                self._errMsg("Unknown database %s: " % database)
                return
            system = result["Value"][database]["System"]
            dbType = result["Value"][database]["Type"]
            setup = gConfig.getValue("/DIRAC/Setup", "")
            if not setup:
                self._errMsg("Unknown current setup")
                return
            instance = gConfig.getValue("/DIRAC/Setups/%s/%s" % (setup, system), "")
            if not instance:
                self._errMsg("No instance defined for system %s" % system)
                self._errMsg("\tAdd new instance with 'add instance %s <instance_name>'" % system)
                return

            if dbType == "MySQL":
                if not gComponentInstaller.mysqlPassword:
                    gComponentInstaller.mysqlPassword = "LocalConfig"
                gComponentInstaller.getMySQLPasswords()
                result = client.installDatabase(database, gComponentInstaller.mysqlRootPwd)
                if not result["OK"]:
                    self._errMsg(result["Message"])
                    return
                extension, system = result["Value"]

                result = client.getHostInfo()
                if not result["OK"]:
                    self._errMsg(result["Message"])
                    return
                else:
                    cpu = result["Value"]["CPUModel"]
                hostname = self.host
                if not result["OK"]:
                    self._errMsg(result["Message"])
                    return

                if database != "InstalledComponentsDB":
                    result = MonitoringUtilities.monitorInstallation(
                        "DB", system.replace("System", ""), database, cpu=cpu, hostname=hostname
                    )
                    if not result["OK"]:
                        self._errMsg(result["Message"])
                        return

                gComponentInstaller.mysqlHost = self.host
                result = client.getInfo()
                if not result["OK"]:
                    self._errMsg(result["Message"])
                hostSetup = result["Value"]["Setup"]

            result = gComponentInstaller.addDatabaseOptionsToCS(gConfig, system, database, hostSetup, overwrite=True)
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            gLogger.notice("Database %s from %s/%s installed successfully" % (database, extension, system))

        # DIRAC components
        elif option in self.runitComponents:
            if len(argss) < 2:
                gLogger.notice(self.do_install.__doc__)
                return

            system = argss[0]
            del argss[0]
            component = argss[0]
            del argss[0]

            specialOptions = {}
            module = ""

            for i in range(len(argss)):
                if argss[i] == "-m":
                    specialOptions["Module"] = argss[i + 1]
                    module = argss[i + 1]
                if argss[i] == "-p":
                    opt, value = argss[i + 1].split("=")
                    specialOptions[opt] = value
            if module == component:
                module = ""

            client = SystemAdministratorClient(self.host, self.port)
            # First need to update the CS
            # result = client.addDefaultOptionsToCS( option, system, component )
            gComponentInstaller.host = self.host
            result = client.getInfo()
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            hostSetup = result["Value"]["Setup"]

            # Install Module section if not yet there
            if module:
                result = gComponentInstaller.addDefaultOptionsToCS(
                    gConfig, option, system, module, extensionsByPriority(), hostSetup
                )
                # in case of Error we must stop, this can happen when the module name is wrong...
                if not result["OK"]:
                    self._errMsg(result["Message"])
                    return
                # Add component section with specific parameters only
                result = gComponentInstaller.addDefaultOptionsToCS(
                    gConfig,
                    option,
                    system,
                    component,
                    extensionsByPriority(),
                    hostSetup,
                    specialOptions,
                    addDefaultOptions=True,
                )
            else:
                # Install component section
                result = gComponentInstaller.addDefaultOptionsToCS(
                    gConfig, option, system, component, extensionsByPriority(), hostSetup, specialOptions
                )

            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            # Then we can install and start the component
            result = client.setupComponent(option, system, component, module)
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            compType = result["Value"]["ComponentType"]
            runit = result["Value"]["RunitStatus"]
            gLogger.notice("%s %s_%s is installed, runit status: %s" % (compType, system, component, runit))

            # And register it in the database
            result = client.getHostInfo()
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            else:
                cpu = result["Value"]["CPUModel"]
            hostname = self.host
            if component == "ComponentMonitoring":
                # Make sure that the service is running before trying to use it
                nTries = 0
                maxTries = 5
                mClient = ComponentMonitoringClient()
                result = mClient.ping()
                while not result["OK"] and nTries < maxTries:
                    time.sleep(3)
                    result = mClient.ping()
                    nTries = nTries + 1

                if not result["OK"]:
                    self._errMsg(
                        "ComponentMonitoring service taking too long to start, won't be logged into the database"
                    )
                    return

                result = MonitoringUtilities.monitorInstallation(
                    "DB", system, "InstalledComponentsDB", cpu=cpu, hostname=hostname
                )
                if not result["OK"]:
                    self._errMsg("Error registering installation into database: %s" % result["Message"])
                    return

            result = MonitoringUtilities.monitorInstallation(
                option, system, component, module, cpu=cpu, hostname=hostname
            )
            if not result["OK"]:
                self._errMsg("Error registering installation into database: %s" % result["Message"])
                return
        else:
            gLogger.notice("Unknown option:", option)

    def do_uninstall(self, args):
        """
        Uninstall a DIRAC component or host along with all its components

        usage:

          uninstall db <database>
          uninstall host <hostname>
          uninstall <-f ForceLogUninstall> <system> <component>
        """
        argss = args.split()
        if not argss:
            gLogger.notice(self.do_uninstall.__doc__)
            return

        # Retrieve user uninstalling the component
        result = getProxyInfo()
        if not result["OK"]:
            self._errMsg(result["Message"])

        option = argss[0]
        if option == "db":
            del argss[0]
            if not argss:
                gLogger.notice(self.do_uninstall.__doc__)
                return
            component = argss[0]
            client = SystemAdministratorClient(self.host, self.port)

            result = client.getHostInfo()
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            else:
                cpu = result["Value"]["CPUModel"]
            hostname = self.host
            result = client.getAvailableDatabases()
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            system = result["Value"][component]["System"]
            result = MonitoringUtilities.monitorUninstallation(system, component, hostname=hostname, cpu=cpu)
            if not result["OK"]:
                self._errMsg(result["Message"])
                return

            result = client.uninstallDatabase(component)
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                gLogger.notice("Successfully uninstalled %s" % (component))
        elif option == "host":
            del argss[0]
            if not argss:
                gLogger.notice(self.do_uninstall.__doc__)
                return
            hostname = argss[0]

            client = ComponentMonitoringClient()
            result = client.hostExists({"HostName": hostname})
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                if not result["Value"]:
                    self._errMsg("Given host does not exist")
                else:
                    result = client.getHosts({"HostName": hostname}, True, False)
                    if not result["OK"]:
                        self._errMsg(result["Message"])
                    else:
                        host = result["Value"][0]
                        # Remove every installation associated with the host
                        for installation in host["Installations"]:
                            result = client.removeInstallations(installation, {}, {"HostName": hostname})
                            if not result["OK"]:
                                self._errMsg(result["Message"])
                                break
                        # Finally remove the host
                        result = client.removeHosts({"HostName": hostname})
                        if not result["OK"]:
                            self._errMsg(result["Message"])
                        else:
                            gLogger.notice("Host %s was successfully removed" % hostname)
        else:
            if option == "-f":
                force = True
                del argss[0]
            else:
                force = False

            if len(argss) != 2:
                gLogger.notice(self.do_uninstall.__doc__)
                return

            system, component = argss
            client = SystemAdministratorClient(self.host, self.port)

            monitoringClient = ComponentMonitoringClient()
            result = monitoringClient.getInstallations(
                {"Instance": component, "UnInstallationTime": None},
                {"DIRACSystem": system},
                {"HostName": self.host},
                True,
            )
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            if len(result["Value"]) < 1:
                self._errMsg("Given component does not exist")
                return
            if len(result["Value"]) > 1:
                self._errMsg("Too many components match")
                return

            removeLogs = False
            if force:
                removeLogs = True
            else:
                if result["Value"][0]["Component"]["Type"] in self.runitComponents:
                    result = promptUser("Remove logs?", ["y", "n"], "n")
                    if result["OK"]:
                        removeLogs = result["Value"] == "y"

            result = client.uninstallComponent(system, component, removeLogs)
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                gLogger.notice("Successfully uninstalled %s/%s" % (system, component))

            result = client.getHostInfo()
            if not result["OK"]:
                self._errMsg(result["Message"])
                return
            else:
                cpu = result["Value"]["CPUModel"]
            hostname = self.host
            result = MonitoringUtilities.monitorUninstallation(system, component, hostname=hostname, cpu=cpu)
            if not result["OK"]:
                return result

    def do_start(self, args):
        """
        Start services or agents or database server

        usage:

          start <system|*> <service|agent|*>
          start mysql
        """
        argss = args.split()
        if len(argss) < 2:
            gLogger.notice(self.do_start.__doc__)
            return
        option = argss[0]
        del argss[0]

        if option != "mysql":
            if len(argss) < 1:
                gLogger.notice(self.do_start.__doc__)
                return
            system = option
            if system != "*":
                component = argss[0]
            else:
                component = "*"
            client = SystemAdministratorClient(self.host, self.port)
            result = client.startComponent(system, component)
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                if system != "*" and component != "*":
                    gLogger.notice("\n%s_%s started successfully, runit status:\n" % (system, component))
                else:
                    gLogger.notice("\nComponents started successfully, runit status:\n")
                for comp in result["Value"]:
                    gLogger.notice((comp.rjust(32), ":", result["Value"][comp]["RunitStatus"]))
        else:
            gLogger.notice("Not yet implemented")

    def do_restart(self, args):
        """
        Restart services or agents or database server

        usage:

          restart <system|*> <service|agent|*>
          restart mysql
        """
        if not args:
            gLogger.notice(self.do_restart.__doc__)
            return

        argss = args.split()
        option = argss[0]
        del argss[0]
        if option != "mysql":
            if option != "*":
                if len(argss) < 1:
                    gLogger.notice(self.do_restart.__doc__)
                    return
            system = option
            if system != "*":
                component = argss[0]
            else:
                component = "*"
            client = SystemAdministratorClient(self.host, self.port)
            result = client.restartComponent(system, component)
            if not result["OK"]:
                if system == "*":
                    gLogger.notice("All systems are restarted, connection to SystemAdministrator is lost")
                else:
                    self._errMsg(result["Message"])
            else:
                if system != "*" and component != "*":
                    gLogger.notice("\n%s_%s started successfully, runit status:\n" % (system, component))
                else:
                    gLogger.notice("\nComponents started successfully, runit status:\n")
                for comp in result["Value"]:
                    gLogger.notice((comp.rjust(32), ":", result["Value"][comp]["RunitStatus"]))
        else:
            gLogger.notice("Not yet implemented")

    def do_stop(self, args):
        """
        Stop services or agents or database server

        usage:

          stop <system|*> <service|agent|*>
          stop mysql
        """
        if not args:
            gLogger.notice(self.do_stop.__doc__)
            return

        argss = args.split()
        if argss[0] != "mysql":
            system = argss[0]
            if system != "*":
                component = argss[1]
            else:
                component = "*"
            client = SystemAdministratorClient(self.host, self.port)
            result = client.stopComponent(system, component)
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                if system != "*" and component != "*":
                    gLogger.notice("\n%s_%s stopped successfully, runit status:\n" % (system, component))
                else:
                    gLogger.notice("\nComponents stopped successfully, runit status:\n")
                for comp in result["Value"]:
                    gLogger.notice((comp.rjust(32), ":", result["Value"][comp]["RunitStatus"]))
        else:
            gLogger.notice("Not yet implemented")

    def do_update(self, args):
        """
        Update the software on the target host to a given version

        usage:

          update <version> [ -g <diracOSVersion> ]

              where lcgVersion - version of DIRACOS to install
        """
        try:
            argss = args.split()
            version = argss[0]
            diracOSVersion = ""
            del argss[0]

            while len(argss) > 0:
                if argss[0] == "-g":
                    diracOSVersion = argss[1]
                    del argss[0]
                    del argss[0]
        except Exception as x:
            gLogger.notice("ERROR: wrong input:", str(x))
            gLogger.notice(self.do_update.__doc__)
            return

        client = SystemAdministratorClient(self.host, self.port)
        gLogger.notice("Software update can take a while, please wait ...")
        result = client.updateSoftware(version, "", diracOSVersion, timeout=300)
        if not result["OK"]:
            self._errMsg("Failed to update the software")
            gLogger.notice(result["Message"])
        else:
            gLogger.notice("Software successfully updated.")
            gLogger.notice("You should restart the services to use the new software version.")
            gLogger.notice("Think of updating /Operations/<vo>/<setup>/Pilot/Versions section in the CS")

    def do_revert(self, args):
        """
        Revert the last installed version of software to the previous one

        usage:

            revert
        """
        client = SystemAdministratorClient(self.host, self.port)
        result = client.revertSoftware()
        if not result["OK"]:
            gLogger.notice("Error:", result["Message"])
        else:
            gLogger.notice("Software reverted to", result["Value"])

    def do_add(self, args):
        """
        Add new entity to the Configuration Service

        usage:

          add system <system> <instance>
        """
        if not args:
            gLogger.notice(self.do_add.__doc__)
            return

        argss = args.split()
        option = argss[0]
        del argss[0]
        if option == "instance" or option == "system":
            system = argss[0]
            instance = argss[1]
            client = SystemAdministratorClient(self.host, self.port)
            result = client.getInfo()
            if not result["OK"]:
                self._errMsg(result["Message"])
            hostSetup = result["Value"]["Setup"]
            instanceName = gConfig.getValue("/DIRAC/Setups/%s/%s" % (hostSetup, system), "")
            if instanceName:
                if instanceName == instance:
                    gLogger.notice(
                        "System %s already has instance %s defined in %s Setup" % (system, instance, hostSetup)
                    )
                else:
                    self._errMsg(
                        "System %s already has instance %s defined in %s Setup" % (system, instance, hostSetup)
                    )
                return
            result = gComponentInstaller.addSystemInstance(system, instance, hostSetup)
            if not result["OK"]:
                self._errMsg(result["Message"])
            else:
                gLogger.notice("%s system instance %s added successfully" % (system, instance))
        else:
            gLogger.notice("Unknown option:", option)

    def do_exec(self, args):
        """
        Execute a shell command on the remote host and get back the output

        usage:

          exec <cmd> [<arguments>]
        """
        client = SystemAdministratorClient(self.host, self.port)
        command = "cd %s;" % self.cwd + args
        result = client.executeCommand(command)
        if not result["OK"]:
            self._errMsg(result["Message"])
            return
        status, output, error = result["Value"]
        gLogger.notice("")
        for line in output.split("\n"):
            gLogger.notice(line)
        if error:
            self._errMsg(status)
            for line in error.split("\n"):
                gLogger.notice(line)

    def do_cd(self, args):
        """
        Change the current working directory on the target host

        Usage:
          cd <dirpath>
        """
        argss = args.split()

        if len(argss) == 0:
            # Return to $HOME
            if self.homeDir:
                self.previous_cwd = self.cwd
                self.cwd = self.homeDir
            else:
                client = SystemAdministratorClient(self.host, self.port)
                command = "echo $HOME"
                result = client.executeCommand(command)
                if not result["OK"]:
                    self._errMsg(result["Message"])
                    return
                status, output, _error = result["Value"]
                if not status and output:
                    self.homeDir = output.strip()
                    self.previous_cwd = self.cwd
                    self.cwd = self.homeDir
            self.prompt = "[%s:%s]> " % (self.host, self.cwd)
            return

        newPath = argss[0]
        if newPath == "-":
            if self.previous_cwd:
                cwd = self.cwd
                self.cwd = self.previous_cwd
                self.previous_cwd = cwd
        elif newPath.startswith("/"):
            self.previous_cwd = self.cwd
            self.cwd = newPath
        else:
            newPath = self.cwd + "/" + newPath
            self.previous_cwd = self.cwd
            self.cwd = os.path.normpath(newPath)
        self.prompt = "[%s:%s]> " % (self.host, self.cwd)

    def do_showall(self, args):
        """
        Show status of all the components in all the hosts

        Usage:
          showall [-snmth] [-ASE] [-N name] [-H host] - show status of components

        Options:
            -d extra debug printout
          Sorting options:
            -s system
            -n component name
            -m component module
            -t component type
            -h component host
          Selection options:
            -A select agents
            -S select services
            -E select executors
            -N <component pattern> select component with the name containing the pattern
            -H <host name> select the given host
            -T <setup name> select the given setup
        """

        argss = args.split()
        sortOption = ""
        componentType = ""
        componentName = ""
        hostName = ""
        setupName = ""
        debug = False
        while len(argss) > 0:
            option = argss[0]
            del argss[0]
            sortOption = ""
            if option == "-s":
                sortOption = "System"
            elif option == "-n":
                sortOption = "Name"
            elif option == "-m":
                sortOption = "Module"
            elif option == "-t":
                sortOption = "Type"
            elif option == "-h":
                sortOption = "Host"
            elif option == "-A":
                componentType = "Agents"
            elif option == "-S":
                componentType = "Services"
            elif option == "-E":
                componentType = "Executors"
            elif option == "-d":
                debug = True
            elif option == "-N":
                componentName = argss[0]
                del argss[0]
            elif option == "-H":
                hostName = argss[0]
                del argss[0]
            elif option == "-T":
                setupName = argss[0]
                del argss[0]
            else:
                self._errMsg("Invalid option %s" % option)
                return

        client = SystemAdministratorIntegrator()
        silentHosts = client.getSilentHosts()
        respondingHosts = client.getRespondingHosts()
        totalHosts = len(silentHosts) + len(respondingHosts)
        resultAll = client.getOverallStatus()
        resultInfo = client.getInfo()

        if not resultAll["OK"]:
            self._errMsg(resultAll["Message"])
        else:
            fields = ["System", "Name", "Module", "Type", "Setup", "Host", "Runit", "Uptime"]
            records = []
            for host in resultAll["Value"]:
                if hostName and hostName not in host:
                    continue
                result = resultAll["Value"][host]
                if not result["OK"]:
                    if debug:
                        self._errMsg("Host %s: %s" % (host, result["Message"]))
                    continue
                rDict = result["Value"]
                for compType in rDict:
                    if componentType and componentType != compType:
                        continue
                    for system in rDict[compType]:
                        components = sorted(rDict[compType][system])
                        for component in components:
                            if componentName and componentName not in component:
                                continue
                            record = []
                            if rDict[compType][system][component]["Installed"]:
                                module = str(rDict[compType][system][component]["Module"])
                                record += [system, component, module, compType.lower()[:-1]]
                                if resultInfo["OK"] and host in resultInfo["Value"] and resultInfo["Value"][host]["OK"]:
                                    setup = resultInfo["Value"][host]["Value"]["Setup"]
                                else:
                                    setup = "Unknown"
                                if setupName and setupName not in setup:
                                    continue
                                record += [setup]
                                record += [host]
                                record += [str(rDict[compType][system][component]["RunitStatus"])]
                                record += [str(rDict[compType][system][component]["Timeup"])]
                                records.append(record)
            printTable(fields, records, sortOption)
            if silentHosts:
                print("\n %d out of %d hosts did not respond" % (len(silentHosts), totalHosts))

    def default(self, args):

        argss = args.split()
        command = argss[0]
        if command in ["ls", "cat", "pwd", "chown", "chmod", "chgrp", "id", "date", "uname", "cp", "mv", "scp"]:
            self.do_exec(args)



"""
   OpenStackEndpoint is Endpoint base class implementation for the OpenStack cloud service.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

__RCSID__ = "$Id$"

import requests
import json
import base64

# DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Cloud.Endpoint import Endpoint
from DIRAC.Resources.Cloud.KeystoneClient import KeystoneClient
from DIRAC.Core.Utilities.File import makeGuid

DEBUG = False


class OpenStackEndpoint(Endpoint):
    """OpenStack implementation of the Cloud Endpoint interface"""

    def __init__(self, parameters=None, bootstrapParameters=None):
        super(OpenStackEndpoint, self).__init__(parameters=parameters, bootstrapParameters=bootstrapParameters)
        # logger
        self.log = gLogger.getSubLogger("OpenStackEndpoint")
        self.ks = None
        self.flavors = {}
        self.images = {}
        self.networks = {}
        self.computeURL = None
        self.imageURL = None
        self.networkURL = None
        self.network = None
        self.project = None
        self.projectID = None
        self.vmInfo = {}
        self.initialized = False

        result = self.initialize()
        if result["OK"]:
            self.log.debug("OpenStackEndpoint created and validated")
        else:
            self.log.error(result["Message"])

    def initialize(self):
        self.caPath = self.parameters.get("CAPath", True)
        self.network = self.parameters.get("Network")
        self.project = self.parameters.get("Project")
        keyStoneURL = self.parameters.get("AuthURL")
        result = self.getProxyFileLocation()
        if result["OK"]:
            self.parameters["Proxy"] = result["Value"]
        self.ks = KeystoneClient(keyStoneURL, self.parameters)
        result = self.ks.getToken()
        if not result["OK"]:
            return result
        self.valid = True
        self.token = result["Value"]
        self.computeURL = self.ks.computeURL
        self.imageURL = self.ks.imageURL
        self.networkURL = self.ks.networkURL
        self.projectID = self.ks.projectID

        self.log.verbose(
            "Service interfaces:\ncompute %s,\nimage %s,\nnetwork %s"
            % (self.computeURL, self.imageURL, self.networkURL)
        )

        result = self.getFlavors()
        if not result["OK"]:
            self.valid = False
        result = self.getImages()
        if not result["OK"]:
            self.valid = False
        self.getNetworks()
        return result

    def getFlavors(self):

        if not self.computeURL or not self.token:
            return S_ERROR("The endpoint object is not initialized")

        url = "%s/flavors/detail" % self.computeURL
        self.log.verbose("Getting flavors details on %s" % url)

        result = requests.get(url, headers={"X-Auth-Token": self.token}, verify=self.caPath)

        output = json.loads(result.text)
        for flavor in output["flavors"]:
            self.flavors[flavor["name"]] = {
                "FlavorID": flavor["id"],
                "RAM": flavor["ram"],
                "NumberOfProcessors": flavor["vcpus"],
            }

        return S_OK(self.flavors)

    def getImages(self):

        if not self.imageURL or not self.token:
            return S_ERROR("The endpoint object is not initialized")

        result = requests.get("%s/v2/images" % self.imageURL, headers={"X-Auth-Token": self.token}, verify=self.caPath)

        output = json.loads(result.text)
        for image in output["images"]:
            self.images[image["name"]] = {"id": image["id"]}

        return S_OK(self.images)

    def getNetworks(self):
        """Get a network object corresponding to the networkName

        :param str networkName: network name
        :return: S_OK|S_ERROR network object in case of S_OK
        """
        try:
            result = requests.get(
                "%s/v2.0/networks" % self.networkURL, headers={"X-Auth-Token": self.token}, verify=self.caPath
            )
            output = json.loads(result.text)
        except Exception as exc:
            return S_ERROR("Cannot get networks: %s" % str(exc))

        for network in output["networks"]:
            if network["project_id"] == self.projectID:
                self.networks[network["name"]] = {"NetworkID": network["id"]}
        return S_OK(self.networks)

    def createInstances(self, vmsToSubmit):
        outputDict = {}
        for nvm in range(vmsToSubmit):
            instanceID = makeGuid()[:8]
            result = self.createInstance(instanceID)
            if result["OK"]:
                nodeID = result["Value"]
                self.log.debug("Created VM instance %s/%s" % (nodeID, instanceID))
                nodeDict = {}
                nodeDict["InstanceID"] = instanceID
                nodeDict["NumberOfProcessors"] = self.parameters["NumberOfProcessors"]
                outputDict[nodeID] = nodeDict
            else:
                self.log.error("Failed to create OpenStack instance", "%s %s %s" % (nvm, instanceID, result["Message"]))
                break

        # We failed submission utterly
        if not outputDict:
            return S_ERROR("No VM submitted")

        return S_OK(outputDict)

    def createInstance(self, instanceID=""):
        """
        This creates a VM instance for the given boot image
        and creates a context script, taken the given parameters.
        Successful creation returns instance VM

        Boots a new node on the OpenStack server defined by self.endpointConfig. The
        'personality' of the node is done by self.imageConfig. Both variables are
        defined on initialization phase.

        The node name has the following format:
        <bootImageName><contextMethod><time>

        :return: S_OK( ( nodeID, publicIP ) ) | S_ERROR
        """

        if not self.initialized:
            self.initialize()

        imageID = self.parameters.get("ImageID")
        if not imageID:
            imageName = self.parameters.get("Image")
            if not imageName:
                return S_ERROR("No image name or ID is specified")
            if not self.images:
                result = self.getImages()
                if not result["OK"]:
                    return result
            imageID = self.images.get(imageName)["id"]
            if not imageID:
                return S_ERROR("Can not get ID for the image: %s" % imageName)
        self.parameters["ImageID"] = imageID
        if "Image" not in self.parameters:
            for image in self.images:
                if self.images[image]["id"] == imageID:
                    self.parameters["Image"] = image

        flavorID = self.parameters.get("FlavorID")
        if not flavorID:
            flavor = self.parameters.get("FlavorName")
            if not flavor:
                return S_ERROR("No flavor name or ID is specified")
            if not self.flavors:
                result = self.getFlavors()
                if not result["OK"]:
                    return result
            flavorID = self.flavors.get(flavor)["FlavorID"]
            if not flavorID:
                return S_ERROR("Can not get ID for the flavor: %s" % flavor)
            numberOfProcessors = self.flavors.get(flavor)["NumberOfProcessors"]
        self.parameters["FlavorID"] = flavorID
        if "NumberOfProcessors" not in self.parameters:
            self.parameters["NumberOfProcessors"] = numberOfProcessors

        networkID = self.parameters.get("NetworkID")
        if not networkID:
            network = self.parameters.get("Network")
            if not self.networks and self.networkURL:
                result = self.getNetworks()
                if not result["OK"]:
                    return result
            if network:
                if network in self.networks:
                    networkID = self.networks[network]["NetworkID"]
            elif self.networks:
                randomNW = self.networks.keys()[0]
                networkID = self.networks[randomNW]["NetworkID"]
            if not networkID:
                self.log.warn("Failed to get ID of the network interface")

        # Pass VMUUID to the instance
        self.parameters["VMUUID"] = instanceID

        result = self._createUserDataScript()
        if not result["OK"]:
            return result
        userDataCrude = str(result["Value"])
        userData = base64.b64encode(userDataCrude.encode()).decode()

        headers = {"X-Auth-Token": self.token}
        requestDict = {
            "server": {
                "user_data": userData,
                "name": "DIRAC_%s" % instanceID,
                "imageRef": imageID,
                "flavorRef": flavorID,
            }
        }
        # Some cloud sites do not expose network service interface, but some do
        if networkID:
            requestDict["server"]["networks"] = [{"uuid": networkID}]

        # Allow the use of pre-uploaded SSH keys
        osSSHKey = self.parameters.get("OSKeyName")
        if osSSHKey:
            requestDict["server"]["key_name"] = osSSHKey

        # print "AT >>> user data", userDataCrude
        # print "AT >>> requestDict", requestDict
        # return S_ERROR()

        try:
            result = requests.post(
                "%s/servers" % self.computeURL, json=requestDict, headers=headers, verify=self.caPath
            )
        except Exception as exc:
            self.log.exception("Exception creating VM")
            return S_ERROR("Exception creating VM: %s" % str(exc))

        if result.status_code in [200, 201, 202, 203, 204]:
            output = json.loads(result.text)
            nodeID = output["server"]["id"]
            return S_OK(nodeID)
        else:
            return S_ERROR("Error creating VM: %s" % result.text)

    def getVMIDs(self):
        """Get all the VM IDs on the endpoint

        :return: list of VM ids
        """

        if not self.initialized:
            result = self.initialize()
            if not result["OK"]:
                return result

        try:
            response = requests.get(
                "%s/servers" % self.computeURL, headers={"X-Auth-Token": self.token}, verify=self.caPath
            )
        except Exception as e:
            return S_ERROR("Cannot connect to " + str(self.computeURL) + " (" + str(e) + ")")

        output = json.loads(response.text)
        idList = []
        for server in output["servers"]:
            idList.append(server["id"])
        return S_OK(idList)

    def getVMStatus(self, vmID):
        """
        Get the status for a given node ID. libcloud translates the status into a digit
        from 0 to 4 using a many-to-one relation ( ACTIVE and RUNNING -> 0 ), which
        means we cannot undo that translation. It uses an intermediate states mapping
        dictionary, SITEMAP, which we use here inverted to return the status as a
        meaningful string. The five possible states are ( ordered from 0 to 4 ):
        RUNNING, REBOOTING, TERMINATED, PENDING & UNKNOWN.

        :Parameters:
          **uniqueId** - `string`
            openstack node id ( not uuid ! )

        :return: S_OK( status ) | S_ERROR
        """

        if not self.initialized:
            self.initialize()

        result = self.getVMInfo(vmID)
        if not result["OK"]:
            return result

        output = result["Value"]
        return S_OK(output["server"])

    def stopVM(self, nodeID):
        """
        Given the node ID it gets the node details, which are used to destroy the
        node

        :param str uniqueId:  openstack node id ( not uuid ! )

        :return: S_OK | S_ERROR
        """

        if not self.initialized:
            self.initialize()

        try:
            response = requests.delete(
                "%s/servers/%s" % (self.computeURL, nodeID), headers={"X-Auth-Token": self.token}, verify=self.caPath
            )
        except Exception as e:
            return S_ERROR("Cannot get node details for %s (" % nodeID + str(e) + ")")

        if response.status_code == 204:
            # VM stopped successfully
            return S_OK(response.text)
        elif response.status_code == 404:
            # VM does not exist already
            return S_OK(response.text)
        else:
            return S_ERROR(response.text)

    def __getVMPortID(self, nodeID):
        """Get the port ID associated with the given VM

        :param str nodeID: VM ID
        :return: port ID
        """
        if nodeID in self.vmInfo and "portID" in self.vmInfo[nodeID]:
            return S_OK(self.vmInfo[nodeID]["portID"])

        # Get the port of my VM
        try:
            result = requests.get(
                "%s/v2.0/ports" % self.networkURL, headers={"X-Auth-Token": self.token}, verify=self.caPath
            )
            output = json.loads(result.text)
            portID = None
            for port in output["ports"]:
                if port["device_id"] == nodeID:
                    portID = port["id"]
                    self.vmInfo.setdefault(nodeID, {})
                    self.vmInfo[nodeID]["portID"] = portID
        except Exception as exc:
            return S_ERROR("Cannot get ports: %s" % str(exc))

        return S_OK(portID)

    def assignFloatingIP(self, nodeID):
        """
        Given a node, assign a floating IP from the ipPool defined on the imageConfiguration
        on the CS.

        :Parameters:
          **node** - `libcloud.compute.base.Node`
            node object with the vm details

        :return: S_OK( public_ip ) | S_ERROR
        """

        if not self.initialized:
            self.initialize()

        result = self.getVMFloatingIP(nodeID)
        if result["OK"]:
            ip = result["Value"]
            if ip:
                return S_OK(ip)

        # Get the port of my VM
        result = self.__getVMPortID(nodeID)
        if not result["OK"]:
            return result
        portID = result["Value"]

        # Get an available floating IP
        try:
            result = requests.get(
                "%s/v2.0/floatingips" % self.networkURL, headers={"X-Auth-Token": self.token}, verify=self.caPath
            )
            output = json.loads(result.text)
        except Exception as e:
            return S_ERROR("Cannot get floatingips")

        fipID = None
        for fip in output["floatingips"]:
            if fip["fixed_ip_address"] is None:
                fipID = fip["id"]
                break

        if fipID is None:
            return S_ERROR("No floating IP available")

        data = {"floatingip": {"port_id": portID}}
        dataJson = json.dumps(data)

        try:
            result = requests.put(
                "%s/v2.0/floatingips/%s" % (self.networkURL, fipID),
                data=dataJson,
                headers={"X-Auth-Token": self.token},
                verify=self.caPath,
            )
        except Exception as e:
            return S_ERROR("Cannot assign floating IP")

        output = json.loads(result.text)

        self.vmInfo.setdefault(nodeID, {})
        self.vmInfo["floatingID"] = output["floatingip"]["id"]

        output = json.loads(result.text)
        self.vmInfo.setdefault(nodeID, {})
        self.vmInfo["floatingID"] = output["floatingip"]["id"]

        ip = output["floatingip"]["floating_ip_address"]
        return S_OK(ip)

    def getVMInfo(self, vmID):

        try:
            response = requests.get(
                "%s/servers/%s" % (self.computeURL, vmID), headers={"X-Auth-Token": self.token}, verify=self.caPath
            )
        except Exception as e:
            return S_ERROR("Cannot get node details for %s (" % vmID + str(e) + ")")

        if response.status_code == 404:
            return S_ERROR("VM ID %s not found" % vmID)

        output = json.loads(response.text)
        if response.status_code == 403:
            if "forbidden" in output:
                return S_ERROR("Cannot get VM info: %s" % output["forbidden"].get("message"))
            else:
                return S_ERROR("Cannot get VM info: access forbidden")

        # Cache some info
        if response.status_code == 200:
            self.vmInfo.setdefault(vmID, {})
            self.vmInfo[vmID]["imageID"] = output["server"]["image"]["id"]
            self.vmInfo[vmID]["flavorID"] = output["server"]["flavor"]["id"]

        return S_OK(output)

    def getVMFloatingIP(self, nodeID):

        result = self.getVMInfo(nodeID)
        if not result["OK"]:
            return result

        floatingIP = None
        output = result["Value"]
        for network, addressList in output["server"]["addresses"].items():
            for address in addressList:
                if address["OS-EXT-IPS:type"] == "floating":
                    floatingIP = address["addr"]

        return S_OK(floatingIP)

    def deleteFloatingIP(self, nodeID, floatingIP=None):
        """
        Deletes a floating IP <public_ip> from the server.

        :param str publicIP: public IP to be deleted
        :param object node: node to which IP is attached
        :return: S_OK | S_ERROR
        """

        if nodeID in self.vmInfo and "floatingID" in self.vmInfo[nodeID]:
            fipID = self.vmInfo[nodeID]["floatingID"]
        else:
            result = self.getVMFloatingIP(nodeID)
            if not result["OK"]:
                return result
            ip = result["Value"]
            if ip is None:
                return S_OK()

            result = self.__getVMPortID(nodeID)
            if not result["OK"]:
                return result

            portID = result["Value"]
            # Get an available floating IP
            try:
                result = requests.get(
                    "%s/v2.0/floatingips" % self.networkURL, headers={"X-Auth-Token": self.token}, verify=self.caPath
                )
                output = json.loads(result.text)
            except Exception as e:
                return S_ERROR("Cannot get floatingips")

            fipID = None
            for fip in output["floatingips"]:
                if fip["port_id"] == portID:
                    fipID = fip["id"]
                    break

        if not fipID:
            return S_ERROR("Can not get the floating IP ID")

        data = {"floatingip": {"port_id": None}}
        dataJson = json.dumps(data)

        try:
            result = requests.put(
                "%s/v2.0/floatingips/%s" % (self.networkURL, fipID),
                data=dataJson,
                headers={"X-Auth-Token": self.token},
                verify=self.caPath,
            )
        except Exception as exc:
            return S_ERROR("Cannot disassociate floating IP: %s" % str(exc))

        if result.status_code == 200:
            return S_OK(fipID)
        else:
            return S_ERROR("Cannot disassociate floating IP: %s" % result.text)



#!/usr/bin/env python
"""
Set the status of the replicas of given files at the provided SE
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LFN:      LFN", "File:     File name containing a list of affected LFNs"))
    Script.registerArgument(" SE:       Name of Storage Element")
    Script.registerArgument(" Status:   New Status for the replica")
    Script.parseCommandLine(ignoreErrors=False)

    import DIRAC
    from DIRAC import gLogger
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
    import os

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    inputFileName, storageElement, status = Script.getPositionalArgs(group=True)

    if os.path.exists(inputFileName):
        inputFile = open(inputFileName, "r")
        string = inputFile.read()
        inputFile.close()
        lfns = sorted(string.splitlines())
    else:
        lfns = [inputFileName]

    fc = FileCatalog()

    res = fc.getReplicas(lfns, allStatus=True)
    if not res["OK"]:
        gLogger.error("Failed to get catalog replicas.", res["Message"])
        DIRAC.exit(-1)
    lfnDict = {}
    for lfn, error in res["Value"]["Failed"].items():
        gLogger.error("Failed to get replicas for file.", "%s:%s" % (lfn, error))
    for lfn, replicas in res["Value"]["Successful"].items():
        if storageElement not in replicas.keys():
            gLogger.error("LFN not registered at provided storage element.", "%s %s" % (lfn, storageElement))
        else:
            lfnDict[lfn] = {"SE": storageElement, "PFN": replicas[storageElement], "Status": status}
    if not lfnDict:
        gLogger.error("No files found at the supplied storage element.")
        DIRAC.exit(2)

    res = fc.setReplicaStatus(lfnDict)
    if not res["OK"]:
        gLogger.error("Failed to set catalog replica status.", res["Message"])
        DIRAC.exit(-1)
    for lfn, error in res["Value"]["Failed"].items():
        gLogger.error("Failed to set replica status for file.", "%s:%s" % (lfn, error))
    gLogger.notice(
        "Successfully updated the status of %d files at %s." % (len(res["Value"]["Successful"].keys()), storageElement)
    )
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" The Bdii2CSAgent performs checking BDII for availability of CE
resources for a given or any configured VO. It detects resources not yet
present in the CS and notifies the administrators.
For the CEs already present in the CS, the agent is updating
if necessary settings which were changed in the BDII recently

The following options can be set for the Bdii2CSAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN Bdii2CSAgent
  :end-before: ##END
  :dedent: 2
  :caption: Bdii2CSAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOs, getVOOption
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getQueues, getCESiteMapping
from DIRAC.ConfigurationSystem.Client.Utilities import getGridCEs, getSiteUpdates
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.Glue2 import getGlue2CEInfo
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient


class Bdii2CSAgent(AgentModule):
    def __init__(self, *args, **kwargs):
        """Defines default parameters"""

        super(Bdii2CSAgent, self).__init__(*args, **kwargs)

        self.addressTo = ""
        self.addressFrom = ""
        self.voName = []
        self.subject = self.am_getModuleParam("fullName")
        self.alternativeBDIIs = []
        self.voBdiiCEDict = {}
        self.voBdiiSEDict = {}
        self.host = "cclcgtopbdii01.in2p3.fr:2170"
        self.injectSingleCoreQueues = False
        self.csAPI = None

        # What to get
        self.processCEs = True
        self.selectedSites = []

        # Update the CS or not?
        self.dryRun = False

    def initialize(self):
        """Gets run paramaters from the configuration"""

        self.addressTo = self.am_getOption("MailTo", self.addressTo)
        self.addressFrom = self.am_getOption("MailFrom", self.addressFrom)
        # Create a list of alternative bdii urls
        self.alternativeBDIIs = self.am_getOption("AlternativeBDIIs", self.alternativeBDIIs)
        self.host = self.am_getOption("Host", self.host)
        self.injectSingleCoreQueues = self.am_getOption("InjectSingleCoreQueues", self.injectSingleCoreQueues)

        # Check if the bdii url is appended by a port number, if not append the default 2170
        for index, url in enumerate(self.alternativeBDIIs):
            if not url.split(":")[-1].isdigit():
                self.alternativeBDIIs[index] += ":2170"
        if self.addressTo and self.addressFrom:
            self.log.info("MailTo", self.addressTo)
            self.log.info("MailFrom", self.addressFrom)
        if self.alternativeBDIIs:
            self.log.info("AlternativeBDII URLs:", self.alternativeBDIIs)

        self.processCEs = self.am_getOption("ProcessCEs", self.processCEs)
        self.selectedSites = self.am_getOption("SelectedSites", [])
        self.dryRun = self.am_getOption("DryRun", self.dryRun)

        self.voName = self.am_getOption("VirtualOrganization", self.voName)
        if not self.voName:
            self.voName = self.am_getOption("VO", [])
        if not self.voName or (len(self.voName) == 1 and self.voName[0].lower() == "all"):
            # Get all VOs defined in the configuration
            self.voName = []
            result = getVOs()
            if result["OK"]:
                vos = result["Value"]
                for vo in vos:
                    vomsVO = getVOOption(vo, "VOMSName")
                    if vomsVO:
                        self.voName.append(vomsVO)

        if self.voName:
            self.log.info("Agent will manage VO(s) %s" % self.voName)
        else:
            self.log.fatal("VirtualOrganization option not defined for agent")
            return S_ERROR()

        self.csAPI = CSAPI()
        return self.csAPI.initialize()

    def execute(self):
        """General agent execution method"""
        self.voBdiiCEDict = {}

        # Get a "fresh" copy of the CS data
        result = self.csAPI.downloadCSData()
        if not result["OK"]:
            self.log.warn("Could not download a fresh copy of the CS data", result["Message"])

        # Refresh the configuration from the master server
        gConfig.forceRefresh(fromMaster=True)

        if self.processCEs:
            self.__lookForNewCEs()
            self.__updateCEs()
        return S_OK()

    def __lookForNewCEs(self):
        """Look up BDII for CEs not yet present in the DIRAC CS"""

        bannedCEs = self.am_getOption("BannedCEs", [])

        for vo in self.voName:
            # get the known CEs for a given VO, so we can know the unknowns, or no longer supported,
            # for a VO
            res = getQueues(community=vo)
            if not res["OK"]:
                return res

            knownCEs = set()
            for _site, ces in res["Value"].items():
                knownCEs.update(ces)
            knownCEs.update(bannedCEs)

            result = self.__getGlue2CEInfo(vo)
            if not result["OK"]:
                continue
            bdiiInfo = result["Value"]
            result = getGridCEs(vo, bdiiInfo=bdiiInfo, ceBlackList=knownCEs)
            if not result["OK"]:
                self.log.error("Failed to get unused CEs", result["Message"])
                continue  # next VO
            siteDict = result["Value"]
            unknownCEs = set(result["UnknownCEs"]) - set(bannedCEs)

            body = ""
            for site in siteDict:
                newCEs = set(siteDict[site])  # pylint: disable=no-member
                if not newCEs:
                    continue

                ceString = ""
                for ce in newCEs:
                    queueString = ""
                    ceInfo = bdiiInfo[site]["CEs"][ce]
                    newCEString = "CE: %s, GOCDB Site Name: %s" % (ce, site)
                    systemTuple = siteDict[site][ce]["System"]
                    osString = "%s_%s_%s" % (systemTuple)
                    newCEString = "\n%s\n%s\n" % (newCEString, osString)
                    for queue in ceInfo["Queues"]:
                        queueStatus = ceInfo["Queues"][queue].get("GlueCEStateStatus", "UnknownStatus")
                        if "production" in queueStatus.lower():
                            ceType = ceInfo["Queues"][queue].get("GlueCEImplementationName", "")
                            queueString += "   %s %s %s\n" % (queue, queueStatus, ceType)
                    if queueString:
                        ceString += newCEString
                        ceString += "Queues:\n"
                        ceString += queueString

                if ceString:
                    body += ceString

            if siteDict:
                body = "\nWe are glad to inform You about new CE(s) possibly suitable for %s:\n" % vo + body
                body += "\n\nTo suppress information about CE add its name to BannedCEs list.\n"
                body += "Add new Sites/CEs for vo %s with the command:\n" % vo
                body += "dirac-admin-add-resources --vo %s --ce\n" % vo

            if unknownCEs:
                body += "\n\n"
                body += "There is no (longer) information about the following CEs for the %s VO.\n" % vo
                body += "\n".join(sorted(unknownCEs))
                body += "\n\n"

            if body:
                self.log.info(body)
                if self.addressTo and self.addressFrom:
                    notification = NotificationClient()
                    result = notification.sendMail(
                        self.addressTo, self.subject, body, self.addressFrom, localAttempt=False
                    )
                    if not result["OK"]:
                        self.log.error("Can not send new site notification mail", result["Message"])

        return S_OK()

    def __getGlue2CEInfo(self, vo):

        if vo in self.voBdiiCEDict:
            return S_OK(self.voBdiiCEDict[vo])
        self.log.info("Check for available CEs for VO", vo)
        totalResult = S_OK({})
        message = ""

        mainResult = getGlue2CEInfo(vo, host=self.host)
        if not mainResult["OK"]:
            self.log.error("Failed getting information from default bdii", mainResult["Message"])
            message = mainResult["Message"]

        for bdii in reversed(self.alternativeBDIIs):
            resultAlt = getGlue2CEInfo(vo, host=bdii)
            if resultAlt["OK"]:
                totalResult["Value"].update(resultAlt["Value"])
            else:
                self.log.error("Failed getting information from %s " % bdii, resultAlt["Message"])
                message = (message + "\n" + resultAlt["Message"]).strip()

        if mainResult["OK"]:
            totalResult["Value"].update(mainResult["Value"])

        if not totalResult["Value"] and message:  # Dict is empty and we have an error message
            self.log.error("Error during BDII request", message)
            totalResult = S_ERROR(message)
        else:
            self.voBdiiCEDict[vo] = totalResult["Value"]
            self.__purgeSites(totalResult["Value"])

        return totalResult

    def __updateCEs(self):
        """Update the Site/CE/queue settings in the CS if they were changed in the BDII"""

        bdiiChangeSet = set()
        bannedCEs = self.am_getOption("BannedCEs", [])

        for vo in self.voName:
            result = self.__getGlue2CEInfo(vo)
            if not result["OK"]:
                continue
            ceBdiiDict = result["Value"]

            for _siteName, ceDict in ceBdiiDict.items():
                for bannedCE in bannedCEs:
                    ceDict["CEs"].pop(bannedCE, None)

            result = getSiteUpdates(vo, bdiiInfo=ceBdiiDict, log=self.log, onecore=self.injectSingleCoreQueues)

            if not result["OK"]:
                continue
            bdiiChangeSet = bdiiChangeSet.union(result["Value"])

        # We have collected all the changes, consolidate VO settings
        result = self.__updateCS(bdiiChangeSet)
        return result

    def __purgeSites(self, ceBdiiDict):
        """Remove all sites that are not in self.selectedSites.

        Modifies the ceBdiiDict!
        """
        if not self.selectedSites:
            return
        for site in list(ceBdiiDict):
            ces = list(ceBdiiDict[site]["CEs"])
            if not ces:
                self.log.error("No CE information for site:", site)
                continue
            siteInCS = "Not_In_CS"
            for ce in ces:
                res = getCESiteMapping(ce)
                if not res["OK"]:
                    self.log.error("Failed to get DIRAC site name for ce", "%s: %s" % (ce, res["Message"]))
                    continue
                # if the ce is not in the CS the returned value will be empty
                if ce in res["Value"]:
                    siteInCS = res["Value"][ce]
                    break
            self.log.debug("Checking site %s (%s), aka %s" % (site, ces, siteInCS))
            if siteInCS in self.selectedSites:
                continue
            self.log.info("Dropping site %s, aka %s" % (site, siteInCS))
            ceBdiiDict.pop(site)
        return

    def __updateCS(self, bdiiChangeSet):

        queueVODict = {}
        changeSet = set()
        for entry in bdiiChangeSet:
            section, option, _value, new_value = entry
            if option == "VO":
                queueVODict.setdefault(section, set())
                queueVODict[section] = queueVODict[section].union(set(new_value.split(",")))
            else:
                changeSet.add(entry)
        for section, VOs in queueVODict.items():  # can be an iterator
            changeSet.add((section, "VO", "", ",".join(VOs)))

        if changeSet:
            changeList = sorted(changeSet)
            body = "\n".join(["%s/%s %s -> %s" % entry for entry in changeList])
            if body and self.addressTo and self.addressFrom:
                notification = NotificationClient()
                result = notification.sendMail(self.addressTo, self.subject, body, self.addressFrom, localAttempt=False)

            if body:
                self.log.info("The following configuration changes were detected:")
                self.log.info(body)

            for section, option, value, new_value in changeSet:
                if value == "Unknown" or not value:
                    self.csAPI.setOption(cfgPath(section, option), new_value)
                else:
                    self.csAPI.modifyValue(cfgPath(section, option), new_value)

            if self.dryRun:
                self.log.info("Dry Run: CS won't be updated")
                self.csAPI.showDiff()
            else:
                result = self.csAPI.commit()
                if not result["OK"]:
                    self.log.error("Error while committing to CS", result["Message"])
                else:
                    self.log.info("Successfully committed %d changes to CS" % len(changeList))
                return result
        else:
            self.log.info("No changes found")
            return S_OK()



"""
    This module implements the default behavior for the FTS3 system for TPC and source SE selection
"""

import random
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers


class DefaultFTS3Plugin(object):
    """ "
    Default FTS3 plugin.

    For the TPC selection, it returns the list configured in the CS.
    For the source SE selection, it calls
    :py:func:`DIRAC.DataManagementSystem.private.FTS3Utilities.selectUniqueRandomSource`

    It is used to document what are the requirements for a real TPC plugin.
    It is a good idea for your plugin to inherit from this one if you only want
    to change one specific behavior.

    Such plugins are meant to alter the TPC protocols list that an FTS3 job
    will use to transfer between two SEs, and possibly make a smart selection
    of the source SE.

    They are called by :py:class:`DIRAC.DataManagementSystem.Client.FTS3Operation.FTS3Operation`

    The class name must be "<PluginName>FTS3Plugin" """

    def __init__(self, vo=None):
        """The plugin is instanciated once per ``FTS3Operation``, so it is a
        good place to do global initialization

        :param str vo: Virtual Organization
        """
        self.vo = vo
        self.thirdPartyProtocols = DMSHelpers(vo=vo).getThirdPartyProtocols()

    def selectTPCProtocols(self, ftsJob=None, sourceSEName=None, destSEName=None, **kwargs):
        """
        This method has to return an ordered list of protocols
        that will be used by the source/dest StorageElements to agree
        on a common TPC protocol.

        There are two ways to invoke the plugin. Either with an FTS3Job instance, or
        with specific parameters.

        The FTS3Job object passed as parameter can be used to make a choice
        based on various parameters.

        Specific parameters are passed when there is access to an FTS3Job already,
        like in the ``__needsMultiHopStaging`` function of
        :py:mod:`~DIRAC.DataManagementSystem.Client.FTS3Operation.FTS3Operation`.

        Thus, it is possible that there is not enough information to make a decision without the FTS3Job.
        In that case, it is up to the plugin to decide whether to return the best possible answer or to raise
        a ``ValueError`` exception


        In this default implementation, we just return the preference list
        that we have in the CS

        :param ftsJob: :py:class:`~DIRAC.DataManagementSystem.Client.FTS3Job.FTS3Job` that will submit the transfer
        :param sourceSEName: Name of the source StorageElement
        :param destSEName: Name of the destination StorageElement

        :returns: an ordered TPC protocols list
        :raise ValueError: in case the plugin cannot select a protocol with the given info
        """
        return self.thirdPartyProtocols

    def selectSourceSE(self, ftsFile, replicaDict, allowedSources):
        """
        For a given FTS3file object, select a source.

        Note that the replicaDict may already have been filtered
        (for example, only active replicas are taken into account),
        so if you want to do exotic things, you may want to recheck the
        replicas

        The ``allowedSources`` is what comes from the RMS, so possibly
        from the TS. So up to you to ignore it or not

        In this default implementation, we only consider the allowed sources
        and the active replicas, preferably on disk (already filtered in replicaDict)
        and return a random choice

        :param ftsFiles: list of FTS3File object
        :param replicaDict: list of replicas for the file
        :param allowedSources: list of allowed sources

        :return:  one SE name
        :raise ValueError: in case the plugin cannot select a sourceSE
        """
        allowedSourcesSet = set(allowedSources) if allowedSources else set()
        # Only consider the allowed sources

        # If we have a restriction, apply it, otherwise take all the replicas
        allowedReplicaSource = (set(replicaDict) & allowedSourcesSet) if allowedSourcesSet else replicaDict

        if not allowedReplicaSource:
            raise ValueError("No valid replicas")

        # pick a random source

        randSource = random.choice(list(allowedReplicaSource))  # one has to convert to list
        return randSource

    def inferFTSActivity(self, ftsOperation, rmsRequest, rmsOperation):
        """
        This will try to find which FTS activity should be applied to
        the FTS3Operation.

        If nothing can be found, it will return None.

        Note that this is only called if there is no hardcoded Activity in the
        RMS Operation Arguments.
        """
        return None



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-submit
# Author :  Stuart Paterson
########################################################################
"""
Submit jobs to DIRAC WMS

Example:
  $ dirac-wms-job-submit Simple.jdl
  JobID = 11
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("f:", "File=", "Writes job ids to file <value>")
    Script.registerSwitch("r:", "UseJobRepo=", "Use the job repository")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JDL:    Path to JDL file"])
    sws, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.Dirac import Dirac

    unprocessed_switches = sws
    use_repo = False
    repo_name = ""
    for sw, value in unprocessed_switches:
        if sw.lower() in ["r", "usejobrepo"]:
            use_repo = True
            repo_name = value
            repo_name = repo_name.replace(".cfg", ".repo")
    dirac = Dirac(use_repo, repo_name)
    exitCode = 0
    errorList = []

    jFile = None
    for sw, value in unprocessed_switches:
        if sw.lower() in ("f", "file"):
            if os.path.isfile(value):
                print("Appending job ids to existing logfile: %s" % value)
                if not os.access(value, os.W_OK):
                    print("Existing logfile %s must be writable by user." % value)
            jFile = open(value, "a")

    for jdl in args:

        result = dirac.submitJob(jdl)
        if result["OK"]:
            print("JobID = %s" % (result["Value"]))
            if jFile is not None:
                # parametric jobs
                if isinstance(result["Value"], list):
                    jFile.write("\n".join(str(p) for p in result["Value"]))
                    jFile.write("\n")
                else:
                    jFile.write(str(result["Value"]) + "\n")
        else:
            errorList.append((jdl, result["Message"]))
            exitCode = 2

    if jFile is not None:
        jFile.close()

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" JobStateUpdateHandler is the implementation of the Job State updating
    service in the DISET framework

    The following methods are available in the Service interface

    setJobStatus()

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import six

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.WorkloadManagementSystem.Client import JobStatus


class JobStateUpdateHandlerMixin(object):
    @classmethod
    def initializeHandler(cls, svcInfoDict):
        """
        Determines the switching of ElasticSearch and MySQL backends
        """
        try:
            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.JobDB", "JobDB")
            if not result["OK"]:
                return result
            cls.jobDB = result["Value"]()

            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.JobLoggingDB", "JobLoggingDB")
            if not result["OK"]:
                return result
            cls.jobLoggingDB = result["Value"]()

        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)

        cls.elasticJobParametersDB = None
        useESForJobParametersFlag = Operations().getValue("/Services/JobMonitoring/useESForJobParametersFlag", False)
        if useESForJobParametersFlag:
            try:
                result = ObjectLoader().loadObject(
                    "WorkloadManagementSystem.DB.ElasticJobParametersDB", "ElasticJobParametersDB"
                )
                if not result["OK"]:
                    return result
                cls.elasticJobParametersDB = result["Value"]()
            except RuntimeError as excp:
                return S_ERROR("Can't connect to DB: %s" % excp)
        return S_OK()

    ###########################################################################
    types_updateJobFromStager = [[six.string_types, int], six.string_types]

    @classmethod
    def export_updateJobFromStager(cls, jobID, status):
        """Simple call back method to be used by the stager."""
        if status == JobStatus.DONE:
            jobStatus = JobStatus.CHECKING
            minorStatus = "JobScheduling"
        elif status == JobStatus.FAILED:
            jobStatus = JobStatus.FAILED
            minorStatus = "Staging input files failed"
        else:
            return S_ERROR("updateJobFromStager: %s status not known." % status)

        infoStr = None
        trials = 10
        for i in range(trials):
            result = cls.jobDB.getJobAttributes(jobID, ["Status"])
            if not result["OK"]:
                return result
            if not result["Value"]:
                # if there is no matching Job it returns an empty dictionary
                return S_OK("No Matching Job")
            status = result["Value"]["Status"]
            if status == JobStatus.STAGING:
                if i:
                    infoStr = "Found job in Staging after %d seconds" % i
                break
            time.sleep(1)
        if status != JobStatus.STAGING:
            return S_OK("Job is not in Staging after %d seconds" % trials)

        result = cls.__setJobStatus(int(jobID), status=jobStatus, minorStatus=minorStatus, source="StagerSystem")
        if not result["OK"]:
            if result["Message"].find("does not exist") != -1:
                return S_OK()
        if infoStr:
            return S_OK(infoStr)
        return result

    ###########################################################################
    types_setJobStatus = [[six.string_types, int], six.string_types, six.string_types, six.string_types]

    @classmethod
    def export_setJobStatus(cls, jobID, status="", minorStatus="", source="Unknown", datetime=None, force=False):
        """
        Sets the major and minor status for job specified by its JobId.
        Sets optionally the status date and source component which sends the status information.
        The "force" flag will override the WMS state machine decision.
        """
        return cls.__setJobStatus(
            int(jobID), status=status, minorStatus=minorStatus, source=source, datetime=datetime, force=force
        )

    @classmethod
    def __setJobStatus(
        cls, jobID, status=None, minorStatus=None, appStatus=None, source=None, datetime=None, force=False
    ):
        """update the job provided statuses (major, minor and application)
        If sets also the source and the time stamp (or current time)
        This method calls the bulk method internally
        """
        sDict = {}
        if status:
            sDict["Status"] = status
        if minorStatus:
            sDict["MinorStatus"] = minorStatus
        if appStatus:
            sDict["ApplicationStatus"] = appStatus
        if sDict:
            if source:
                sDict["Source"] = source
            if not datetime:
                datetime = Time.toString()
            return cls.__setJobStatusBulk(jobID, {datetime: sDict}, force=force)
        return S_OK()

    ###########################################################################
    types_setJobStatusBulk = [[six.string_types, int], dict]

    @classmethod
    def export_setJobStatusBulk(cls, jobID, statusDict, force=False):
        """Set various job status fields with a time stamp and a source"""
        return cls.__setJobStatusBulk(jobID, statusDict, force=force)

    @classmethod
    def __setJobStatusBulk(cls, jobID, statusDict, force=False):
        """Set various status fields for job specified by its JobId.
        Set only the last status in the JobDB, updating all the status
        logging information in the JobLoggingDB. The statusDict has datetime
        as a key and status information dictionary as values
        """
        status = ""
        minor = ""
        application = ""
        jobID = int(jobID)

        result = cls.jobDB.getJobAttributes(jobID, ["Status", "StartExecTime", "EndExecTime"])
        if not result["OK"]:
            return result

        if not result["Value"]:
            # if there is no matching Job it returns an empty dictionary
            return S_ERROR("No Matching Job")
        # If the current status is Stalled and we get an update, it should probably be "Running"
        currentStatus = result["Value"]["Status"]
        if currentStatus == JobStatus.STALLED:
            status = JobStatus.RUNNING
        startTime = result["Value"].get("StartExecTime")
        endTime = result["Value"].get("EndExecTime")
        # getJobAttributes only returns strings :(
        if startTime == "None":
            startTime = None
        if endTime == "None":
            endTime = None

        # Get the latest WN time stamps of status updates
        result = cls.jobLoggingDB.getWMSTimeStamps(int(jobID))
        if not result["OK"]:
            return result
        lastTime = max([float(t) for s, t in result["Value"].items() if s != "LastTime"])
        lastTime = Time.toString(Time.fromEpoch(lastTime))

        dates = sorted(statusDict)
        # If real updates, start from the current status
        if dates[0] >= lastTime and not status:
            status = currentStatus
        log = gLogger.getLocalSubLogger("JobStatusBulk/Job-%s" % jobID)
        log.debug("*** New call ***", "Last update time %s - Sorted new times %s" % (lastTime, dates))
        # Remove useless items in order to make it simpler later, although there should not be any
        for sDict in statusDict.values():
            for item in sorted(sDict):
                if not sDict[item]:
                    sDict.pop(item, None)
        # Pick up start and end times from all updates, if they don't exist
        newStat = status
        for date in dates:
            sDict = statusDict[date]
            # This is to recover Matched jobs that set the application status: they are running!
            if sDict.get("ApplicationStatus") and newStat == JobStatus.MATCHED:
                sDict["Status"] = JobStatus.RUNNING
            newStat = sDict.get("Status", newStat)

            # evaluate the state machine
            if not force and newStat:
                res = JobStatus.JobsStateMachine(currentStatus).getNextState(newStat)
                if not res["OK"]:
                    return res
                nextState = res["Value"]

                # If the JobsStateMachine does not accept the candidate, don't update
                if newStat != nextState:
                    log.error(
                        "Job Status Error",
                        "%s can't move from %s to %s: using %s" % (jobID, currentStatus, newStat, nextState),
                    )
                    newStat = nextState
                sDict["Status"] = newStat
                currentStatus = newStat

            if newStat == JobStatus.RUNNING and not startTime:
                # Pick up the start date when the job starts running if not existing
                startTime = date
                log.debug("Set job start time", startTime)
            elif newStat in JobStatus.JOB_FINAL_STATES and not endTime:
                # Pick up the end time when the job is in a final status
                endTime = date
                log.debug("Set job end time", endTime)

        # We should only update the status if its time stamp is more recent than the last update
        if dates[-1] >= lastTime:
            # Get the last status values
            for date in [dt for dt in dates if dt >= lastTime]:
                sDict = statusDict[date]
                log.debug("\t", "Time %s - Statuses %s" % (date, str(sDict)))
                status = sDict.get("Status", status)
                minor = sDict.get("MinorStatus", minor)
                application = sDict.get("ApplicationStatus", application)

            log.debug("Final statuses:", "status '%s', minor '%s', application '%s'" % (status, minor, application))
            attrNames = []
            attrValues = []
            if status:
                attrNames.append("Status")
                attrValues.append(status)
            if minor:
                attrNames.append("MinorStatus")
                attrValues.append(minor)
            if application:
                attrNames.append("ApplicationStatus")
                attrValues.append(application)
            # Here we are forcing the update as it's always updating to the last status
            result = cls.jobDB.setJobAttributes(jobID, attrNames, attrValues, update=True, force=True)
            if not result["OK"]:
                return result

        # Update start and end time if needed
        if endTime:
            result = cls.jobDB.setEndExecTime(jobID, endTime)
            if not result["OK"]:
                return result
        if startTime:
            result = cls.jobDB.setStartExecTime(jobID, startTime)
            if not result["OK"]:
                return result

        # Update the JobLoggingDB records
        for date in dates:
            sDict = statusDict[date]
            status = sDict.get("Status", "idem")
            minor = sDict.get("MinorStatus", "idem")
            application = sDict.get("ApplicationStatus", "idem")
            source = sDict.get("Source", "Unknown")
            result = cls.jobLoggingDB.addLoggingRecord(
                jobID, status=status, minorStatus=minor, applicationStatus=application, date=date, source=source
            )
            if not result["OK"]:
                return result

        return S_OK()

    ###########################################################################
    types_setJobAttribute = [[six.string_types, int], six.string_types, six.string_types]

    @classmethod
    def export_setJobAttribute(cls, jobID, attribute, value):
        """Set a job attribute"""
        return cls.jobDB.setJobAttribute(int(jobID), attribute, value)

    ###########################################################################
    types_setJobSite = [[six.string_types, int], six.string_types]

    @classmethod
    def export_setJobSite(cls, jobID, site):
        """Allows the site attribute to be set for a job specified by its jobID."""
        return cls.jobDB.setJobAttribute(int(jobID), "Site", site)

    ###########################################################################
    types_setJobFlag = [[six.string_types, int], six.string_types]

    @classmethod
    def export_setJobFlag(cls, jobID, flag):
        """Set job flag for job with jobID"""
        return cls.jobDB.setJobAttribute(int(jobID), flag, "True")

    ###########################################################################
    types_unsetJobFlag = [[six.string_types, int], six.string_types]

    @classmethod
    def export_unsetJobFlag(cls, jobID, flag):
        """Unset job flag for job with jobID"""
        return cls.jobDB.setJobAttribute(int(jobID), flag, "False")

    ###########################################################################
    types_setJobApplicationStatus = [[six.string_types, int], six.string_types, six.string_types]

    @classmethod
    def export_setJobApplicationStatus(cls, jobID, appStatus, source="Unknown"):
        """Set the application status for job specified by its JobId.
        Internally calling the bulk method
        """
        return cls.__setJobStatus(jobID, appStatus=appStatus, source=source)

    ###########################################################################
    types_setJobParameter = [[six.string_types, int], six.string_types, six.string_types]

    @classmethod
    def export_setJobParameter(cls, jobID, name, value):
        """Set arbitrary parameter specified by name/value pair
        for job specified by its JobId
        """

        if cls.elasticJobParametersDB:
            return cls.elasticJobParametersDB.setJobParameter(int(jobID), name, value)  # pylint: disable=no-member

        return cls.jobDB.setJobParameter(int(jobID), name, value)

    ###########################################################################
    types_setJobsParameter = [dict]

    @classmethod
    @ignoreEncodeWarning
    def export_setJobsParameter(cls, jobsParameterDict):
        """Set arbitrary parameter specified by name/value pair
        for job specified by its JobId
        """
        failed = False

        for jobID in jobsParameterDict:

            if cls.elasticJobParametersDB:
                res = cls.elasticJobParametersDB.setJobParameter(
                    jobID, str(jobsParameterDict[jobID][0]), str(jobsParameterDict[jobID][1])
                )
                if not res["OK"]:
                    gLogger.error("Failed to add Job Parameter to elasticJobParametersDB", res["Message"])
                    failed = True
                    message = res["Message"]

            else:
                res = cls.jobDB.setJobParameter(
                    jobID, str(jobsParameterDict[jobID][0]), str(jobsParameterDict[jobID][1])
                )
                if not res["OK"]:
                    gLogger.error("Failed to add Job Parameter to MySQL", res["Message"])
                    failed = True
                    message = res["Message"]

        if failed:
            return S_ERROR(message)
        return S_OK()

    ###########################################################################
    types_setJobParameters = [[six.string_types, int], list]

    @classmethod
    @ignoreEncodeWarning
    def export_setJobParameters(cls, jobID, parameters):
        """Set arbitrary parameters specified by a list of name/value pairs
        for job specified by its JobId
        """
        if cls.elasticJobParametersDB:
            result = cls.elasticJobParametersDB.setJobParameters(jobID, parameters)
            if not result["OK"]:
                gLogger.error("Failed to add Job Parameters to ElasticJobParametersDB", result["Message"])
        else:
            result = cls.jobDB.setJobParameters(int(jobID), parameters)
            if not result["OK"]:
                gLogger.error("Failed to add Job Parameters to MySQL", result["Message"])

        return result

    ###########################################################################
    types_sendHeartBeat = [[six.string_types, int], dict, dict]

    @classmethod
    def export_sendHeartBeat(cls, jobID, dynamicData, staticData):
        """Send a heart beat sign of life for a job jobID"""

        result = cls.jobDB.setHeartBeatData(int(jobID), dynamicData)
        if not result["OK"]:
            gLogger.warn("Failed to set the heart beat data", "for job %d " % int(jobID))

        if cls.elasticJobParametersDB:
            for key, value in staticData.items():
                result = cls.elasticJobParametersDB.setJobParameter(int(jobID), key, value)
                if not result["OK"]:
                    gLogger.error("Failed to add Job Parameters to ElasticSearch", result["Message"])
        else:
            result = cls.jobDB.setJobParameters(int(jobID), list(staticData.items()))
            if not result["OK"]:
                gLogger.error("Failed to add Job Parameters to MySQL", result["Message"])

        # Restore the Running status if necessary
        result = cls.jobDB.getJobAttributes(jobID, ["Status"])
        if not result["OK"]:
            return result

        if not result["Value"]:
            return S_ERROR("Job %d not found" % jobID)

        status = result["Value"]["Status"]
        if status in (JobStatus.STALLED, JobStatus.MATCHED):
            result = cls.jobDB.setJobAttribute(jobID=jobID, attrName="Status", attrValue=JobStatus.RUNNING, update=True)
            if not result["OK"]:
                gLogger.warn("Failed to restore the job status to Running")

        jobMessageDict = {}
        result = cls.jobDB.getJobCommand(int(jobID))
        if result["OK"]:
            jobMessageDict = result["Value"]

        if jobMessageDict:
            for key, _value in jobMessageDict.items():
                result = cls.jobDB.setJobCommandStatus(int(jobID), key, "Sent")

        return S_OK(jobMessageDict)


class JobStateUpdateHandler(JobStateUpdateHandlerMixin, RequestHandler):
    pass



"""
Test Logger Wrapper
"""

__RCSID__ = "$Id$"

import logging
from six import StringIO

from DIRAC.FrameworkSystem.private.standardLogging.LoggingRoot import LoggingRoot
from DIRAC.FrameworkSystem.private.standardLogging.Logging import Logging


gLogger = LoggingRoot()


def cleaningLog(log):
    """
    Remove date and space from the log string
    """
    log = log.split("UTC ")[-1]
    return log


def captureBackend():
    """
    Dirac logger is wrapped by LoggingRoot and represent the root of the DIRAC logging system
    Modify the output to capture logs of LoggingRoot
    """
    bufferDirac = StringIO()
    if logging.getLogger("dirac").handlers:
        logging.getLogger("dirac").handlers[0].stream = bufferDirac
    return bufferDirac


def gLoggerReset():
    """
    Reinitialize gLogger as only one instance exists
    It avoids any unexpected behaviour due to multiple different usages
    """
    # Reinitialize the system/component name after other tests
    # because LoggingRoot is a singleton and can not be reinstancied
    Logging._componentName = "Framework"

    # reset gLogger
    gLogger.setLevel("notice")
    gLogger.showHeaders(True)
    gLogger.showThreadIDs(False)
    gLogger.showContexts(True)
    gLogger.showTimeStamps(True)

    # modify the output to capture the log records into a buffer
    bufferDirac = captureBackend()

    del logging.getLogger("dirac").handlers[1:]
    del gLogger._backendsList[1:]

    # reset log
    logging.getLogger("dirac").getChild("log").setLevel(logging.NOTSET)
    log = gLogger.getSubLogger("log")

    log.showHeaders(True)
    log.showThreadIDs(False)
    log.showContexts(True)
    log.showTimeStamps(True)
    for option in log._optionsModified:
        log._optionsModified[option] = False

    del logging.getLogger("dirac.log").handlers[:]
    del log._backendsList[:]

    # reset sublog
    logging.getLogger("dirac.log").getChild("sublog").setLevel(logging.NOTSET)
    sublog = log.getSubLogger("sublog")

    sublog.showHeaders(True)
    sublog.showThreadIDs(False)
    sublog.showContexts(True)
    sublog.showTimeStamps(True)
    for option in sublog._optionsModified:
        sublog._optionsModified[option] = False

    del logging.getLogger("dirac.log.sublog").handlers[:]
    del sublog._backendsList[:]

    return (bufferDirac, log, sublog)



########################################################################
# File :   PoolComputingElement.py
# Author : A.T.
########################################################################

""" The Pool Computing Element is an "inner" CE (meaning it's used by a jobAgent inside a pilot)

    It's used running several jobs simultaneously in separate processes, managed by a ProcessPool
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import concurrent.futures

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.private.ConfigurationData import ConfigurationData

from DIRAC.Resources.Computing.ComputingElement import ComputingElement

from DIRAC.Resources.Computing.InProcessComputingElement import InProcessComputingElement
from DIRAC.Resources.Computing.SudoComputingElement import SudoComputingElement
from DIRAC.Resources.Computing.SingularityComputingElement import SingularityComputingElement

# Number of unix users to run job payloads with sudo
MAX_NUMBER_OF_SUDO_UNIX_USERS = 32


def executeJob(executableFile, proxy, taskID, **kwargs):
    """wrapper around ce.submitJob: decides which CE to use (Sudo or InProcess or Singularity)

    :param str executableFile: location of the executable file
    :param str proxy: proxy file location to be used for job submission
    :param int taskID: local task ID of the PoolCE

    :return: the result of the job submission (S_OK/S_ERROR)
    """

    innerCESubmissionType = kwargs.pop("InnerCESubmissionType")

    if innerCESubmissionType == "Sudo":
        ce = SudoComputingElement("Task-" + str(taskID))
        payloadUser = kwargs.get("PayloadUser")
        if payloadUser:
            ce.setParameters({"PayloadUser": payloadUser})
    elif innerCESubmissionType == "Singularity":
        ce = SingularityComputingElement("Task-" + str(taskID))
    else:
        ce = InProcessComputingElement("Task-" + str(taskID))

    return ce.submitJob(executableFile, proxy)


class PoolComputingElement(ComputingElement):

    #############################################################################
    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(PoolComputingElement, self).__init__(ceUniqueID)

        self.ceType = "Pool"
        self.submittedJobs = 0
        self.processors = 1
        self.pPool = None
        self.taskID = 0
        self.processorsPerTask = {}
        self.userNumberPerTask = {}

        # This CE will effectively submit to another "Inner"CE
        # (by default to the InProcess CE)
        self.innerCESubmissionType = "InProcess"

    def _reset(self):
        """Update internal variables after some extra parameters are added

        :return: None
        """

        self.processors = int(self.ceParameters.get("NumberOfProcessors", self.processors))
        self.ceParameters["MaxTotalJobs"] = self.processors
        self.innerCESubmissionType = self.ceParameters.get("InnerCESubmissionType", self.innerCESubmissionType)
        return S_OK()

    def getProcessorsInUse(self):
        """Get the number of currently allocated processor cores

        :return: number of processors in use
        """
        processorsInUse = 0
        for future in self.processorsPerTask:
            processorsInUse += self.processorsPerTask[future]
        return processorsInUse

    #############################################################################
    def submitJob(self, executableFile, proxy=None, **kwargs):
        """Method to submit job.
        This method will submit to a ProcessPoolExecutor, which returns Future objects.

        :param str executableFile: location of the executable file
        :param str proxy: payload proxy

        :return: S_OK/S_ERROR of the result of the job submission
        """

        if self.pPool is None:
            self.pPool = concurrent.futures.ProcessPoolExecutor(max_workers=self.processors)

        processorsForJob = self._getProcessorsForJobs(kwargs)
        if not processorsForJob:
            return S_ERROR("Not enough processors for the job")

        # Now persisting the job limits for later use in pilot.cfg file (pilot 3 default)
        cd = ConfigurationData(loadDefaultCFG=False)
        res = cd.loadFile("pilot.cfg")
        if not res["OK"]:
            self.log.error("Could not load pilot.cfg", res["Message"])
        # only NumberOfProcessors for now, but RAM (or other stuff) can also be added
        jobID = int(kwargs.get("jobDesc", {}).get("jobID", 0))
        cd.setOptionInCFG("/Resources/Computing/JobLimits/%d/NumberOfProcessors" % jobID, processorsForJob)
        res = cd.dumpLocalCFGToFile("pilot.cfg")
        if not res["OK"]:
            self.log.error("Could not dump cfg to pilot.cfg", res["Message"])

        # Here we define task kwargs: adding complex objects like thread.Lock can trigger errors in the task
        taskKwargs = {"InnerCESubmissionType": self.innerCESubmissionType}
        if self.innerCESubmissionType == "Sudo":
            for nUser in range(MAX_NUMBER_OF_SUDO_UNIX_USERS):
                if nUser not in self.userNumberPerTask.values():
                    break
            taskKwargs["NUser"] = nUser
            if "USER" in os.environ:
                taskKwargs["PayloadUser"] = os.environ["USER"] + "p%s" % str(nUser).zfill(2)

        future = self.pPool.submit(executeJob, executableFile, proxy, self.taskID, **taskKwargs)
        self.processorsPerTask[future] = processorsForJob
        self.taskID += 1
        future.add_done_callback(self.finalizeJob)

        return S_OK()  # returning S_OK as callback will do the rest

    def _getProcessorsForJobs(self, kwargs):
        """helper function"""
        processorsInUse = self.getProcessorsInUse()
        availableProcessors = self.processors - processorsInUse

        self.log.verbose(
            "Processors (total, in use, available)",
            "(%d, %d, %d)" % (self.processors, processorsInUse, availableProcessors),
        )

        # Does this ask for MP?
        if not kwargs.get("mpTag", False):
            if availableProcessors:
                return 1
            else:
                return 0

        # From here we assume the job is asking for MP
        if kwargs.get("wholeNode", False):
            if processorsInUse > 0:
                return 0
            else:
                return self.processors

        if "numberOfProcessors" in kwargs:
            requestedProcessors = int(kwargs["numberOfProcessors"])
        else:
            requestedProcessors = 1

        if availableProcessors < requestedProcessors:
            return 0

        # If there's a maximum number of processors allowed for the job, use that as maximum,
        # otherwise it will use all the remaining processors
        if "maxNumberOfProcessors" in kwargs and kwargs["maxNumberOfProcessors"]:
            requestedProcessors = min(int(kwargs["maxNumberOfProcessors"]), availableProcessors)

        return requestedProcessors

    def finalizeJob(self, future):
        """Finalize the job by updating the process utilisation counters

        :param future: evaluating the future result
        """
        nProc = self.processorsPerTask.pop(future)

        result = future.result()  # This would be the result of the e.g. InProcess.submitJob()
        if result["OK"]:
            self.log.info("Task %s finished successfully, %d processor(s) freed" % (future, nProc))
        else:
            self.log.error("Task failed submission", "%d, message: %s" % (future, result["Message"]))
        self.taskResults[future] = result

    def getCEStatus(self):
        """Method to return information on running and waiting jobs,
            as well as the number of processors (used, and available).

        :return: dictionary of numbers of jobs per status and processors (used, and available)
        """

        result = S_OK()
        nJobs = 0
        for _j, value in self.processorsPerTask.items():
            if value > 0:
                nJobs += 1
        result["SubmittedJobs"] = nJobs
        result["RunningJobs"] = nJobs
        result["WaitingJobs"] = 0

        # dealing with processors
        processorsInUse = self.getProcessorsInUse()
        result["UsedProcessors"] = processorsInUse
        result["AvailableProcessors"] = self.processors - processorsInUse
        return result

    def getDescription(self):
        """Get a list of CEs descriptions (each is a dict)

        This is called by the JobAgent.
        """
        result = super(PoolComputingElement, self).getDescription()
        if not result["OK"]:
            return result
        ceDict = result["Value"]

        ceDictList = []
        if self.ceParameters.get("MultiProcessorStrategy"):
            strategyRequiredTags = []
            if not ceDict.get("ProcessorsInUse", 0):
                # We are starting from a clean page, try to get the most demanding
                # jobs first
                strategyRequiredTags.append(["WholeNode"])
            processors = ceDict.get("NumberOfProcessors", 0)
            if processors > 1:
                # We have several processors at hand, try to use most of them
                strategyRequiredTags.append(["%dProcessors" % processors])
                # Well, at least jobs with some processors requirement
                strategyRequiredTags.append(["MultiProcessor"])

            for strat in strategyRequiredTags:
                newCEDict = dict(ceDict)
                newCEDict.setdefault("RequiredTag", []).extend(strat)
                ceDictList.append(newCEDict)

        # Do not require anything special if nothing else was lucky
        ceDictList.append(dict(ceDict))

        return S_OK(ceDictList)

    def shutdown(self):
        """Wait for all futures (jobs) to complete"""
        if self.pPool:
            self.pPool.shutdown()  # blocking
        return S_OK(self.taskResults)



########################################################################
# File :   CREAMComputingElement.py
# Author : A.T.
########################################################################

""" CREAM Computing Element
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os
import re
import tempfile
import stat

from DIRAC import S_OK, S_ERROR

from DIRAC.Resources.Computing.ComputingElement import ComputingElement
from DIRAC.Core.Utilities.Grid import executeGridCommand
from DIRAC.Core.Utilities.File import makeGuid
from DIRAC.WorkloadManagementSystem.Client import PilotStatus


CE_NAME = "CREAM"
MANDATORY_PARAMETERS = ["Queue"]
STATES_MAP = {
    "DONE-OK": PilotStatus.DONE,
    "DONE-FAILED": PilotStatus.FAILED,
    "REGISTERED": PilotStatus.WAITING,
    "PENDING": PilotStatus.WAITING,
    "IDLE": PilotStatus.WAITING,
    "ABORTED": PilotStatus.ABORTED,
    "CANCELLED": PilotStatus.ABORTED,
    "RUNNING": PilotStatus.RUNNING,
    "REALLY-RUNNING": PilotStatus.RUNNING,
    "N/A": PilotStatus.UNKNOWN,
}


class CREAMComputingElement(ComputingElement):

    #############################################################################
    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(CREAMComputingElement, self).__init__(ceUniqueID)

        self.ceType = CE_NAME
        self.submittedJobs = 0
        self.mandatoryParameters = MANDATORY_PARAMETERS
        self.pilotProxy = ""
        self.queue = ""
        self.outputURL = "gsiftp://localhost"
        self.gridEnv = ""
        self.proxyRenewal = 0

    #############################################################################
    def _addCEConfigDefaults(self):
        """Method to make sure all necessary Configuration Parameters are defined"""
        # First assure that any global parameters are loaded
        ComputingElement._addCEConfigDefaults(self)

    def __writeJDL(self, executableFile, processors=1):
        """Create the JDL for submission"""

        workingDirectory = self.ceParameters["WorkingDirectory"]
        fd, name = tempfile.mkstemp(suffix=".jdl", prefix="CREAM_", dir=workingDirectory)
        diracStamp = os.path.basename(name).replace(".jdl", "").replace("CREAM_", "")

        extraJDLParameters = []
        if processors != 1:
            if processors <= 0:
                extraJDLParameters.append("HostNumber = 1")
                extraJDLParameters.append("WholeNodes = true")
            else:
                extraJDLParameters.append("SMPGranularity = %d" % processors)
                extraJDLParameters.append("CPUNumber = %d" % processors)
                extraJDLParameters.append("WholeNodes = false")

        extraParams = self.ceParameters.get("ExtraJDLParameters")
        if extraParams:
            extraJDLParameters += extraParams.strip().split(";")

        extraJDLParameterList = ";\n  ".join([item.strip() for item in extraJDLParameters])

        jdlFile = os.fdopen(fd, "w")

        jdl = """
[
  JobType = "Normal";
  Executable = "%(executable)s";
  StdOutput="%(diracStamp)s.out";
  StdError="%(diracStamp)s.err";
  InputSandbox={"%(executableFile)s"};
  OutputSandbox={"%(diracStamp)s.out", "%(diracStamp)s.err"};
  OutputSandboxBaseDestUri="%(outputURL)s";
  %(extraJDLParameters)s
]
    """ % {
            "executableFile": executableFile,
            "executable": os.path.basename(executableFile),
            "outputURL": self.outputURL,
            "diracStamp": diracStamp,
            "extraJDLParameters": extraJDLParameterList,
        }

        jdlFile.write(jdl)
        jdlFile.close()
        return name, diracStamp

    def _reset(self):
        self.queue = self.ceParameters.get("CEQueueName", self.ceParameters["Queue"])
        self.outputURL = self.ceParameters.get("OutputURL", "gsiftp://localhost")
        self.gridEnv = self.ceParameters.get("GridEnv", self.gridEnv)
        return S_OK()

    #############################################################################
    def submitJob(self, executableFile, proxy, numberOfJobs=1):
        """Method to submit job"""

        self.log.verbose("Executable file path: %s" % executableFile)
        if not os.access(executableFile, 5):
            os.chmod(executableFile, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)

        nProcessors = self.ceParameters.get("NumberOfProcessors", 1)

        batchIDList = []
        stampDict = {}
        if numberOfJobs == 1:
            jdlName, diracStamp = self.__writeJDL(executableFile, processors=nProcessors)
            cmd = ["glite-ce-job-submit", "-n", "-a", "-N", "-r", "%s/%s" % (self.ceName, self.queue), "%s" % jdlName]

            result = executeGridCommand(self.proxy, cmd, self.gridEnv)
            os.unlink(jdlName)
            if result["OK"]:
                if result["Value"][0]:
                    # We have got a non-zero status code
                    errorString = "\n".join(result["Value"][1:]).strip()
                    return S_ERROR("Pilot submission failed with error: %s " % errorString)
                pilotJobReference = result["Value"][1].strip()
                if not pilotJobReference:
                    return S_ERROR("No pilot reference returned from the glite job submission command")
                if not pilotJobReference.startswith("https"):
                    return S_ERROR("Invalid pilot reference %s" % pilotJobReference)
                batchIDList.append(pilotJobReference)
                stampDict[pilotJobReference] = diracStamp
        else:
            delegationID = makeGuid()
            cmd = ["glite-ce-delegate-proxy", "-e", "%s" % self.ceName, "%s" % delegationID]
            result = executeGridCommand(self.proxy, cmd, self.gridEnv)
            if not result["OK"]:
                self.log.error("Failed to delegate proxy", result["Message"])
                return result
            for _i in range(numberOfJobs):
                jdlName, diracStamp = self.__writeJDL(executableFile, processors=nProcessors)
                cmd = [
                    "glite-ce-job-submit",
                    "-n",
                    "-N",
                    "-r",
                    "%s/%s" % (self.ceName, self.queue),
                    "-D",
                    "%s" % delegationID,
                    "%s" % jdlName,
                ]
                result = executeGridCommand(self.proxy, cmd, self.gridEnv)
                os.unlink(jdlName)
                if not result["OK"]:
                    self.log.error("General error in execution of glite-ce-job-submit command")
                    break
                if result["Value"][0] != 0:
                    self.log.error("Error in glite-ce-job-submit command", result["Value"][1] + result["Value"][2])
                    break
                pilotJobReference = result["Value"][1].strip()
                if pilotJobReference and pilotJobReference.startswith("https"):
                    batchIDList.append(pilotJobReference)
                    stampDict[pilotJobReference] = diracStamp
                else:
                    break
        if batchIDList:
            result = S_OK(batchIDList)
            result["PilotStampDict"] = stampDict
        else:
            result = S_ERROR("No pilot references obtained from the glite job submission")
        return result

    def killJob(self, jobIDList):
        """Kill the specified jobs"""
        jobList = list(jobIDList)
        if isinstance(jobIDList, six.string_types):
            jobList = [jobIDList]

        cmd = ["glite-ce-job-cancel", "-n", "-N"] + jobList
        result = executeGridCommand(self.proxy, cmd, self.gridEnv)
        if not result["OK"]:
            return result
        if result["Value"][0] != 0:
            errorString = "\n".join(result["Value"][1:]).strip()
            return S_ERROR("Failed kill job: %s" % errorString)

        return S_OK()

    #############################################################################
    def getCEStatus(self, jobIDList=None):
        """Method to return information on running and pending jobs.

        :param jobIDList: list of job IDs to be considered
        :type jobIDList: python:list
        """
        statusList = ["REGISTERED", "PENDING", "IDLE", "RUNNING", "REALLY-RUNNING"]
        cmd = ["glite-ce-job-status", "-n", "-a", "-e", "%s" % self.ceName, "-s", "%s" % ":".join(statusList)]
        result = executeGridCommand(self.proxy, cmd, self.gridEnv)
        resultDict = {}
        if not result["OK"]:
            return result
        if result["Value"][0]:
            if result["Value"][0] == 11:
                return S_ERROR("Segmentation fault while calling glite-ce-job-status")
            elif result["Value"][2]:
                return S_ERROR(result["Value"][2])
            elif "Authorization error" in result["Value"][1]:
                return S_ERROR("Authorization error")
            elif "FaultString" in result["Value"][1]:
                res = re.search(r"FaultString=\[([\w\s]+)\]", result["Value"][1])
                fault = ""
                if res:
                    fault = res.group(1)
                detail = ""
                res = re.search(r"FaultDetail=\[([\w\s]+)\]", result["Value"][1])
                if res:
                    detail = res.group(1)
                    return S_ERROR("Error: %s:%s" % (fault, detail))
            else:
                return S_ERROR("Error while interrogating CE status")
        if result["Value"][1]:
            resultDict = self.__parseJobStatus(result["Value"][1])

        running = 0
        waiting = 0
        statusDict = {}
        for ref, status in resultDict.items():
            if jobIDList is not None and ref not in jobIDList:
                continue
            if status == "Scheduled":
                waiting += 1
            if status == "Running":
                running += 1
            statusDict[ref] = status

        result = S_OK()
        result["RunningJobs"] = running
        result["WaitingJobs"] = waiting
        result["SubmittedJobs"] = 0
        result["JobStatusDict"] = statusDict
        return result

    def getJobStatus(self, jobIDList):
        """Get the status information for the given list of jobs"""
        if self.proxyRenewal % 60 == 0:
            self.proxyRenewal += 1
            statusList = ["REGISTERED", "PENDING", "IDLE", "RUNNING", "REALLY-RUNNING"]
            cmd = [
                "glite-ce-job-status",
                "-L",
                "2",
                "--all",
                "-e",
                "%s" % self.ceName,
                "-s",
                "%s" % ":".join(statusList),
            ]
            result = executeGridCommand(self.proxy, cmd, self.gridEnv)
            if result["OK"]:
                delegationIDs = []
                for line in result["Value"][1].split("\n"):
                    if line.find("Deleg Proxy ID") != -1:
                        delegationID = line.split()[-1].replace("[", "").replace("]", "")
                        if delegationID not in delegationIDs:
                            delegationIDs.append(delegationID)
                if delegationIDs:
                    # Renew proxies in batches to avoid timeouts
                    chunkSize = 10
                    for i in range(0, len(delegationIDs), chunkSize):
                        chunk = delegationIDs[i : i + chunkSize]
                        cmd = ["glite-ce-proxy-renew", "-e", self.ceName]
                        cmd.extend(chunk)
                        self.log.info("Refreshing proxy for:", " ".join(chunk))
                        result = executeGridCommand(self.proxy, cmd, self.gridEnv)
                        if result["OK"]:
                            status, output, error = result["Value"]
                            if status:
                                self.log.error(
                                    "Failed to renew proxy delegation", "Output:\n" + output + "\nError:\n" + error
                                )

        workingDirectory = self.ceParameters["WorkingDirectory"]
        fd, idFileName = tempfile.mkstemp(suffix=".ids", prefix="CREAM_", dir=workingDirectory)
        idFile = os.fdopen(fd, "w")
        idFile.write("##CREAMJOBS##")
        for id_ in jobIDList:
            if ":::" in id_:
                ref, _stamp = id_.split(":::")
            else:
                ref = id_
            idFile.write("\n" + ref)
        idFile.close()

        cmd = ["glite-ce-job-status", "-n", "-i", "%s" % idFileName]
        result = executeGridCommand(self.proxy, cmd, self.gridEnv)
        os.unlink(idFileName)
        resultDict = {}
        if not result["OK"]:
            self.log.error("Failed to get job status", result["Message"])
            return result
        if result["Value"][0]:
            if result["Value"][2]:
                return S_ERROR(result["Value"][2])
            return S_ERROR("Error while interrogating job statuses")
        if result["Value"][1]:
            resultDict = self.__parseJobStatus(result["Value"][1])

        if not resultDict:
            return S_ERROR("No job statuses returned")

        # If CE does not know about a job, set the status to Unknown
        for job in jobIDList:
            if job not in resultDict:
                resultDict[job] = PilotStatus.UNKNOWN

        return S_OK(resultDict)

    def __parseJobStatus(self, output):
        """Parse the output of the glite-ce-job-status"""
        resultDict = {}
        ref = ""
        for line in output.split("\n"):
            if not line:
                continue
            match = re.search(r"JobID=\[(.*)\]", line)
            if match and len(match.groups()) == 1:
                ref = match.group(1)
            match = re.search(r"Status.*\[(.*)\]", line)
            if match and len(match.groups()) == 1:
                creamStatus = match.group(1)
                resultDict[ref] = STATES_MAP.get(creamStatus, PilotStatus.UNKNOWN)

        return resultDict

    def getJobLog(self, jobID):
        """Get pilot job logging info

        :param str jobID: pilot job identifier
        :return: string representing the logging info of a given pilot job
        """
        # pilotRef may integrate the pilot stamp
        # it has to be removed before being passed in parameter
        jobID = jobID.split(":::")[0]
        cmd = ["glite-ce-job-status", "-L", "2", "%s" % jobID]
        ret = executeGridCommand("", cmd, self.gridEnv)
        if not ret["OK"]:
            return ret

        status, output, error = ret["Value"]
        if status:
            return S_ERROR(error)

        return S_OK(output)

    def getJobOutput(self, jobID):
        """Get the specified job standard output and error files. The output is returned
        as strings.
        """
        if jobID.find(":::") != -1:
            pilotRef, stamp = jobID.split(":::")
        else:
            pilotRef = jobID
            stamp = ""
        if not stamp:
            return S_ERROR("Pilot stamp not defined for %s" % pilotRef)

        outURL = self.ceParameters.get("OutputURL", "gsiftp://localhost")
        if outURL == "gsiftp://localhost":
            result = self.__resolveOutputURL(pilotRef)
            if not result["OK"]:
                return result
            outURL = result["Value"]

        outputURL = os.path.join(outURL, "%s.out" % stamp)
        errorURL = os.path.join(outURL, "%s.err" % stamp)
        workingDirectory = self.ceParameters["WorkingDirectory"]
        outFileName = os.path.join(workingDirectory, os.path.basename(outputURL))
        errFileName = os.path.join(workingDirectory, os.path.basename(errorURL))

        cmd = ["globus-url-copy", "%s" % outputURL, "file://%s" % outFileName]
        result = executeGridCommand(self.proxy, cmd, self.gridEnv)
        output = ""
        if result["OK"]:
            if not result["Value"][0]:
                outFile = open(outFileName, "r")
                output = outFile.read()
                outFile.close()
                os.unlink(outFileName)
            elif result["Value"][0] == 1 and "No such file or directory" in result["Value"][2]:
                output = "Standard Output is not available on the CREAM service"
                if os.path.exists(outFileName):
                    os.unlink(outFileName)
                return S_ERROR(output)
            else:
                error = "\n".join(result["Value"][1:])
                return S_ERROR(error)
        else:
            return S_ERROR("Failed to retrieve output for %s" % jobID)

        cmd = ["globus-url-copy", "%s" % errorURL, "%s" % errFileName]
        result = executeGridCommand(self.proxy, cmd, self.gridEnv)
        error = ""
        if result["OK"]:
            if not result["Value"][0]:
                errFile = open(errFileName, "r")
                error = errFile.read()
                errFile.close()
                os.unlink(errFileName)
        elif result["Value"][0] == 1 and "No such file or directory" in result["Value"][2]:
            error = "Standard Error is not available on the CREAM service"
            if os.path.exists(errFileName):
                os.unlink(errFileName)
            return S_ERROR(error)
        else:
            return S_ERROR("Failed to retrieve error for %s" % jobID)

        return S_OK((output, error))

    def __resolveOutputURL(self, pilotRef):
        """Resolve the URL of the pilot output files"""

        cmd = ["glite-ce-job-status", "-L", "2", "%s" % pilotRef, "| grep -i osb"]
        result = executeGridCommand(self.proxy, cmd, self.gridEnv)
        url = ""
        if result["OK"]:
            if not result["Value"][0]:
                output = result["Value"][1]
                for line in output.split("\n"):
                    line = line.strip()
                    if line.find("OSB") != -1:
                        match = re.search(r"\[(.*)\]", line)
                        if match:
                            url = match.group(1)
            if url:
                return S_OK(url)
            return S_ERROR("output URL not found for %s" % pilotRef)
        else:
            return S_ERROR("Failed to retrieve long status for %s" % pilotRef)



""" ClassAd Class - a light purely Python representation of the
    Condor ClassAd library.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"
import six


class ClassAd(object):
    def __init__(self, jdl):
        """ClassAd constructor from a JDL string"""
        self.contents = {}
        result = self.__analyse_jdl(jdl)
        if result:
            self.contents = result

    def __analyse_jdl(self, jdl, index=0):
        """Analyse one [] jdl enclosure"""

        jdl = jdl.strip()

        # Strip all the blanks first
        # temp = jdl.replace(' ','').replace('\n','')
        temp = jdl

        result = {}

        if temp[0] != "[" or temp[-1] != "]":
            print("Invalid JDL: it should start with [ and end with ]")
            return result

        # Parse the jdl string now
        body = temp[1:-1]
        index = 0
        namemode = 1
        valuemode = 0
        while index < len(body):
            if namemode:
                ind = body.find("=", index)
                if ind != -1:
                    name = body[index:ind]
                    index = ind + 1
                    valuemode = 1
                    namemode = 0
                else:
                    break
            elif valuemode:
                ind1 = body.find("[", index)
                ind2 = body.find(";", index)
                if ind1 != -1 and ind1 < ind2:
                    value, newind = self.__find_subjdl(body, ind1)
                elif ind1 == -1 and ind2 == -1:
                    value = body[index:]
                    newind = len(body)
                else:
                    if index == ind2:
                        return {}
                    else:
                        value = body[index:ind2]
                        newind = ind2 + 1

                result[name.strip()] = value.strip().replace("\n", "")
                index = newind
                valuemode = 0
                namemode = 1

        return result

    def __find_subjdl(self, body, index):
        """Find a full [] enclosure starting from index"""
        result = ""
        if body[index] != "[":
            return (result, 0)

        depth = 0
        ind = index
        while depth < 10:
            ind1 = body.find("]", ind + 1)
            ind2 = body.find("[", ind + 1)
            if ind2 != -1 and ind2 < ind1:
                depth += 1
                ind = ind2
            else:
                if depth > 0:
                    depth -= 1
                    ind = ind1
                else:
                    result = body[index : ind1 + 1]
                    if body[ind1 + 1] == ";":
                        return (result, ind1 + 2)
                    return result, 0

        return result, 0

    def insertAttributeInt(self, name, attribute):
        """Insert a named integer attribute"""

        self.contents[name] = str(attribute)

    def insertAttributeBool(self, name, attribute):
        """Insert a named boolean attribute"""

        if attribute:
            self.contents[name] = "true"
        else:
            self.contents[name] = "false"

    def insertAttributeString(self, name, attribute):
        """Insert a named string attribute"""

        self.contents[name] = '"' + str(attribute) + '"'

    def insertAttributeVectorString(self, name, attributelist):
        """Insert a named string list attribute"""

        tmp = ['"' + x + '"' for x in attributelist]
        tmpstr = ",".join(tmp)
        self.contents[name] = "{" + tmpstr + "}"

    def insertAttributeVectorInt(self, name, attributelist):
        """Insert a named string list attribute"""

        tmp = [str(x) for x in attributelist]
        tmpstr = ",".join(tmp)
        self.contents[name] = "{" + tmpstr + "}"

    def insertAttributeVectorStringList(self, name, attributelist):
        """Insert a named list of string lists"""

        listOfLists = []
        for stringList in attributelist:
            # tmp = map ( lambda x : '"' + x + '"', stringList )
            tmpstr = ",".join(stringList)
            listOfLists.append("{" + tmpstr + "}")
        self.contents[name] = "{" + ",".join(listOfLists) + "}"

    def lookupAttribute(self, name):
        """Check the presence of the given attribute"""

        return name in self.contents

    def set_expression(self, name, attribute):
        """Insert a named expression attribute"""

        self.contents[name] = str(attribute)

    def get_expression(self, name):
        """Get expression corresponding to a named attribute"""

        if name in self.contents:
            if isinstance(self.contents[name], six.integer_types):
                return str(self.contents[name])
            return self.contents[name]
        return ""

    def isAttributeList(self, name):
        """Check if the given attribute is of the List type"""
        attribute = self.get_expression(name).strip()
        return attribute.startswith("{")

    def getListFromExpression(self, name):
        """Get a list of strings from a given expression"""

        tempString = self.get_expression(name).strip()
        listMode = False
        if tempString.startswith("{"):
            tempString = tempString[1:-1]
            listMode = True

        tempString = tempString.replace(" ", "").replace("\n", "")
        if tempString.find("{") < 0:
            if not listMode:
                tempString = tempString.replace('"', "")
                if not tempString:
                    return []
                return tempString.split(",")

        resultList = []
        while tempString:
            if tempString.find("{") == 0:
                end = tempString.find("}")
                resultList.append(tempString[: end + 1])
                tempString = tempString[end + 1 :]
                if tempString.startswith(","):
                    tempString = tempString[1:]
            elif tempString.find('"') == 0:
                end = tempString[1:].find('"')
                resultList.append(tempString[1 : end + 1])
                tempString = tempString[end + 2 :]
                if tempString.startswith(","):
                    tempString = tempString[1:]
            else:
                end = tempString.find(",")
                if end < 0:
                    resultList.append(tempString.replace('"', "").replace(" ", ""))
                    break
                else:
                    resultList.append(tempString[:end].replace('"', "").replace(" ", ""))
                    tempString = tempString[end + 1 :]

        return resultList

    def getDictionaryFromSubJDL(self, name):
        """Get a dictionary of the JDL attributes from a subsection"""

        tempList = self.get_expression(name)[1:-1]
        resDict = {}
        for item in tempList.split(";"):
            if len(item.split("=")) == 2:
                resDict[item.split("=")[0].strip()] = item.split("=")[1].strip().replace('"', "")
            else:
                return {}

        return resDict

    def deleteAttribute(self, name):
        """Delete a named attribute"""

        if name in self.contents:
            del self.contents[name]
            return 1
        return 0

    def isOK(self):
        """Check the JDL validity - to be defined"""

        if self.contents:
            return 1
        return 0

    def asJDL(self):
        """Convert the JDL description into a string"""

        result = []
        for name, value in sorted(self.contents.items()):
            if value[0:1] == "{":
                result += [4 * " " + name + " = \n"]
                result += [8 * " " + "{\n"]
                strings = value[1:-1].split(",")
                for st in strings:
                    result += [12 * " " + st.strip() + ",\n"]
                result[-1] = result[-1][:-2]
                result += ["\n" + 8 * " " + "};\n"]
            elif value[0:1] == "[":
                tempad = ClassAd(value)
                tempjdl = tempad.asJDL() + ";"
                lines = tempjdl.split("\n")
                result += [4 * " " + name + " = \n"]
                for line in lines:
                    result += [8 * " " + line + "\n"]

            else:
                result += [4 * " " + name + " = " + str(value) + ";\n"]
        if result:
            result[-1] = result[-1][:-1]
        return "[ \n" + "".join(result) + "\n]"

    def getAttributeString(self, name):
        """Get String type attribute value"""
        value = ""
        if self.lookupAttribute(name):
            value = self.get_expression(name).replace('"', "")
        return value

    def getAttributeInt(self, name):
        """Get Integer type attribute value"""
        value = None
        if self.lookupAttribute(name):
            try:
                value = int(self.get_expression(name).replace('"', ""))
            except Exception:
                value = None
        return value

    def getAttributeBool(self, name):
        """Get Boolean type attribute value"""
        if not self.lookupAttribute(name):
            return False

        value = self.get_expression(name).replace('"', "")
        return value.lower() == "true"

    def getAttributeFloat(self, name):
        """Get Float type attribute value"""
        value = None
        if self.lookupAttribute(name):
            try:
                value = float(self.get_expression(name).replace('"', ""))
            except Exception:
                value = None
        return value

    def getAttributes(self):
        """Get the list of all the attribute names

        :return: list of names as strings
        """
        return list(self.contents)
